import pdb

from mergedeep import merge

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.cli.helpers.handle_config_update_service import (
    context as handle_context, handle_intercept_active_update, handle_policy_update, handle_project_update, handle_scenario_update
) 
from stoobly_agent.app.models.scenario_model import ScenarioModel
from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import mock_policy, mode, record_policy, replay_policy
from stoobly_agent.lib.api.keys.project_key import InvalidProjectKey, ProjectKey
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey

class ConfigsController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # GET /configs/policies
    def policies(self, context):
        settings = Settings.instance()
        active_mode = settings.proxy.intercept.active

        if active_mode in [mode.MOCK, mode.TEST]:
            context.render(
                json = [mock_policy.ALL, mock_policy.FOUND],
                status = 200
            )
        elif active_mode == mode.RECORD:
            context.render(
                json = [record_policy.ALL, record_policy.FOUND, record_policy.NOT_FOUND, record_policy.OVERWRITE],
                status = 200
            )
        elif active_mode == mode.REPLAY:
            context.render(
                json = [replay_policy.ALL],
                status = 200
            )

    # GET /configs
    def show(self, context):
        settings = Settings.instance()

        context.render(
            json = settings.to_dict(),
            status = 200
        )

    # GET /configs/summary
    def summary(self, context):
        settings = Settings.instance()
        proxy = settings.proxy
        intercept_settings = InterceptSettings(settings)

        project_key = intercept_settings.project_key
        project_id = ProjectKey(project_key).id if project_key else None
                
        scenario_key = intercept_settings.scenario_key
        scenario_id = None

        # Check to make sure the scenario still exists
        scenario_uuid = ScenarioKey(scenario_key).id if scenario_key else None

        if scenario_uuid:
            model = self.__scenario_model(settings)
            scenario, status = model.show(scenario_uuid)

            if status == 404:
                intercept_settings.scenario_key = None
                settings.commit()
            elif status == 200:
                scenario_id = scenario['id']

        remote_project_id = self.__remote_project_id(settings)

        modes = [mode.RECORD, mode.MOCK, mode.TEST, mode.REPLAY] if not context.params.get('agent') else [mode.RECORD, mode.MOCK, mode.REPLAY]

        context.render(
            json = {
                'active': intercept_settings.active,
                'mode': intercept_settings.mode,
                'modes': modes,
                'project_id': int(project_id) if project_id != None else None,
                'proxy_url': proxy.url,
                'remote_enabled': settings.cli.features.remote,
                'remote_project_id': remote_project_id,
                'scenario_id': int(scenario_id) if scenario_id != None else None,
            },
            status = 200
        )

    # PUT /configs
    def update(self, context):
        settings = Settings.instance()

        merged_settings = merge(settings.to_dict(), context.params)
        settings.from_dict(merged_settings)

        _handle_context = handle_context()

        handle_intercept_active_update(settings, _handle_context)
        handle_policy_update(settings, _handle_context)
        handle_project_update(settings, _handle_context)
        handle_scenario_update(settings, _handle_context)

        settings.write(settings.to_dict())

        context.render(
            json = merged_settings,
            status = 200
        )

    def __remote_project_id(self, settings: Settings):
        remote_project_key = settings.remote.project_key

        try:
            return ProjectKey(remote_project_key).id if remote_project_key else None
        except InvalidProjectKey:
            pass

    def __scenario_model(self, settings: Settings = None):
        scenario_model = ScenarioModel(settings or Settings.instance())
        return scenario_model


import pdb
import uuid

from stoobly_orator.orm import belongs_to, has_many, has_one 

from stoobly_agent.lib.api.keys.project_key import LOCAL_PROJECT_ID
from stoobly_agent.lib.api.keys.request_key import RequestKey
from stoobly_agent.lib.utils.decode import decode

from .base import Base
from .response import Response

class InvalidScenario(Exception):
  pass

class Request(Base):
  __fillable__ = [
    'body_params_hash', 
    'body_text_hash', 
    'pushed_at',
    'control',
    'headers_hash', 
    'host', 
    'http_version',
    'is_deleted',
    'latency',
    'method', 
    'password',
    'path', 
    'port', 
    'query',
    'query_params_hash', 
    'raw', 
    'response_hash',
    'response_headers_hash',
    'scenario_id',
    'scheme', 
    'starred',
    'status',
    'uuid',
    'user',
  ]
  __primary_key__ = 'id'

  @has_one
  def response(self):
    return Response

  @has_many
  def replayed_responses(self):
    return ReplayedResponse

  @belongs_to
  def scenario(self):
    return Scenario

  def key(self):
    return decode(RequestKey.encode(LOCAL_PROJECT_ID, self.uuid))

  # Override
  def to_dict(self):
    h = super().attributes_to_dict()
    h['key'] = self.key()
    return h

  @property
  def url(self):
    s = self.host

    if self.user:
      if not self.password:
        s = f"{self.user}@{s}"
      else:
        s = f"{self.user}:{self.password}@#{s}"

    if self.scheme and len(self.scheme) > 0:
      s = f"{self.scheme}://{s}"

    if self.port != None:
      if not ((self.scheme == 'https' and self.port == 443) or (self.scheme == 'http' and self.port == 80)):
        s = f"{s}:{self.port}"

    s += self.path

    _query = self.get_raw_attribute('query')
    if _query and len(_query) > 0:
      s = f"{s}?{_query}"

    return s

def handle_creating(request):
  if not hasattr(request, 'uuid') or not request.uuid:
    request.uuid = str(uuid.uuid4())
  else:
    try:
      uuid.UUID(request.uuid)
    except Exception:
      request.uuid = str(uuid.uuid4())

def handle_created(request):
  pass

def handle_saving(request):
  if hasattr(request, 'is_deleted') and request.is_deleted:
    request.scenario_id = None

  if hasattr(request, 'scenario_id') and request.scenario_id:
    try:
      # If set as uuid, convert to id
      uuid.UUID(request.scenario_id) 

      scenario = Scenario.find_by(uuid=request.scenario_id)

      if scenario:
        request.scenario_id = scenario.id
      else:
        raise InvalidScenario()
    except Exception:
      scenario = Scenario.find_by(id=request.scenario_id)

      if scenario:
        request.scenario_id = request.scenario_id
      else:
        raise InvalidScenario()

def handle_saved(request):
  request_before = request.get_original()

  if not request_before.get('scenario_id'):
    scenario = request.scenario
    if scenario:
      scenario.requests_count += 1
      scenario.save()
  else:
    if request_before.get('scenario_id') != request.scenario_id:
      scenario = Scenario.find(request_before.get('scenario_id'))
      if scenario:
        scenario.requests_count -= 1
        scenario.save()

      scenario = request.scenario
      if scenario:
        scenario.requests_count += 1
        scenario.save()

def handle_deleting(request):
  response = request.response

  if response:
    response.delete()

  replayed_responses = request.replayed_responses

  for replayed_response in replayed_responses:
    replayed_response.delete()

def handle_deleted(request):
  request_before = request.get_original()

  if request_before.get('scenario_id'):
    # request.scenario returns a cached version of the scenario
    scenario = Scenario.find(request_before['scenario_id'])

    if scenario:
      scenario.requests_count -= 1
      scenario.save()

Request.saved(handle_saved)
Request.saving(handle_saving)
Request.created(handle_created)
Request.creating(handle_creating)
Request.deleted(handle_deleted)
Request.deleting(handle_deleting)

from .replayed_response import ReplayedResponse
from .scenario import Scenario


import os

from .app import App

class Service():

  def __init__(self, service_name: str, app: App):
    self.__app = app
    self.__service_name = service_name

  @property
  def app(self):
    return self.__app

  @property
  def dir_path(self):
    return os.path.join(self.app.scaffold_namespace_path, self.service_name)

  @property
  def service_name(self):
    return self.__service_name

  def workflow_dir_path(self, workflow_name: str):
    return os.path.join(self.dir_path, workflow_name)


from pathlib import Path
from typing import Dict

import pytest

from stoobly_agent.app.cli.helpers.openapi_endpoint_adapter import (
  OpenApiEndpointAdapter,
)


@pytest.mark.openapi
class TestOpenApiEndpointAdapterMissingInfo():
  @pytest.fixture(scope='class')
  def mock_data_directory_path(self):
    return Path(__file__).parent.parent.parent.parent / 'mock_data'

  @pytest.fixture(scope='class')
  def petstore_file_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "petstore-missing-info.yaml"
    return path

  @pytest.fixture(scope='class')
  def open_api_endpoint_adapter(self):
    adapter = OpenApiEndpointAdapter()
    return adapter

  @pytest.fixture(scope='class')
  def expected_get_pets_endpoint(self) -> Dict:
    return {
      "id": 1,
      "method": "GET",
      "host": "-",
      "match_pattern": "/pets",
      "path": "/pets",
      "port": "0",
      "query_param_names": [
        {
          "endpoint_id": 1,
          "inferred_type": "Integer",
          "is_required": False,
          "is_deterministic": True,
          "name": "limit",
          "query": "limit",
          "query_param_name_id": None,
          "id": 1
        }
      ],
     'path_segment_names': [
        {
          'name': 'pets',
          'type': 'static',
        },
      ],
     'response_header_names': [
        {
          'is_deterministic': True,
          'is_required': False,
          'name': 'x-next',
        },
      ],
      "response_param_names": [
        {
          "endpoint_id": 1,
          "name": "Element",
          "query": "[*]",
          "is_required": False,
          "inferred_type": "Hash",
          "is_deterministic": True,
          "id": 1,
          "response_param_name_id": None
        },
        {
          "endpoint_id": 1,
          "name": "id",
          "query": "[*].id",
          "is_required": True,
          "inferred_type": "Integer",
          "is_deterministic": True,
          "id": 2,
          "response_param_name_id": 1,
          "values": [
            0
          ]
        },
        {
          "endpoint_id": 1,
          "name": "name",
          "query": "[*].name",
          "is_required": True,
          "inferred_type": "String",
          "is_deterministic": True,
          "id": 3,
          "response_param_name_id": 1,
          "values": [
            ""
          ]
        },
        {
          "endpoint_id": 1,
          "name": "tag",
          "query": "[*].tag",
          "is_required": False,
          "inferred_type": "String",
          "is_deterministic": True,
          "id": 4,
          "response_param_name_id": 1
        }
      ],
    }

  def test_adapt_from_file(self, open_api_endpoint_adapter, petstore_file_path, expected_get_pets_endpoint):
    adapter = open_api_endpoint_adapter

    endpoints = adapter.adapt_from_file(petstore_file_path)

    assert len(endpoints) == 1
    get_pets_endpoint = endpoints[0]
    assert get_pets_endpoint == expected_get_pets_endpoint




import pdb

from typing import Callable, List

from stoobly_agent.app.proxy.replay.trace_context import TraceContext
from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias
from stoobly_agent.lib.orm.trace_request import TraceRequest

def iterate_group_by(alias_name: str, trace_context: TraceContext, handler: Callable[[TraceContext], None]):
  trace = trace_context.trace
  trace_aliases = trace.trace_aliases

  # Build map trace_request_id => trace_aliase with trace_request_id
  request_groups = __build_request_groups(trace_aliases)

  # Find all request groups that have alias name as one of the aliases
  request_groups_with_alias_name = __group_by_alias_name(request_groups, alias_name) 

  for _trace_aliases in request_groups_with_alias_name:
    trace = Trace.create()
    trace_request = TraceRequest.create(trace_id=trace.id)

    for trace_alias in _trace_aliases:
      _trace_alias = trace_alias.replicate()
      _trace_alias.assigned_to = None
      _trace_alias.trace_id = trace.id
      _trace_alias.trace_request_id = trace_request.id
      _trace_alias.save()

    trace_context = TraceContext(trace_context.endpoints_resource, trace)

    handler(trace_context)

def __build_request_groups(trace_aliases: List[TraceAlias]):
  '''
  Returns { :traceRequestId => { :aliasName => [] } }
  '''
  request_groups = {}

  for trace_alias in trace_aliases:
    trace_request_id = trace_alias.trace_request_id
    if trace_request_id not in request_groups:
      request_groups[trace_request_id] = {}

    trace_alias_name = trace_alias.name
    request_group = request_groups[trace_request_id]
    if trace_alias_name not in request_group:
      request_group[trace_alias_name] = []
    
    request_group[trace_alias_name].append(trace_alias)

  return request_groups

def __group_by_alias_name(request_groups: dict, alias_name: str):
  '''
  Returns list of lists where each have at most one trace_alias with name == alias_name 
  '''
  request_groups_with_alias_name = []

  # request_groups e.g. 
  # {None: {':endpointId': [<stoobly_agent.lib.orm.trace_alias.TraceAlias object at 0x7f88eedeca90>, <stoobly_agent.lib.orm.trace_alias.TraceAlias object at 0x7f88eedec0a0>]}}
  for trace_request_id, request_group in request_groups.items():
    if alias_name not in request_group:
      continue
    
    # request_group[alias_name] e.g. 
    # [<stoobly_agent.lib.orm.trace_alias.TraceAlias object at 0x7f88eedeca90>, <stoobly_agent.lib.orm.trace_alias.TraceAlias object at 0x7f88eedec0a0>]
    for i, trace_alias in enumerate(request_group[alias_name]):
      group = [trace_alias]

      for _alias_name, trace_aliases in request_group.items():
        if len(trace_aliases) == 0:
          continue

        # We have already added a trace_alias.name == alias_name, skip it
        if _alias_name == alias_name:
          continue
        
        trace_alias = trace_aliases[i % len(trace_aliases)]

        group.append(trace_alias)
    
      request_groups_with_alias_name.append(group)

  return request_groups_with_alias_name


import click

from stoobly_agent.app.cli.helpers import run_command, run_command_with_proxy_export
from stoobly_agent.app.settings import Settings

class ExecDecorator():
  def __init__(self, main):
    self.__main = main

  def decorate(self):
    @self.__main.command(
      help="Run shell command with proxy enabled"
    )
    @click.option('--command', is_flag=True, default=False, help='Read commands from the command_string operand instead of from the standard input.')
    @click.option('--shell', default='sh', help='Shell script to run interpret command(s).')
    @click.argument('file_path')
    def exec(**kwargs):
        is_command = kwargs['command']
        shell = kwargs['shell']
        file_path = kwargs['file_path']

        settings = Settings.instance()
        proxy_url = settings.proxy.url

        if not proxy_url:
            run_command(shell, file_path, is_command)
        else:
            run_command_with_proxy_export(shell, file_path, is_command, proxy_url)


import pdb
import requests

from urllib.parse import parse_qs, urlparse

from .mitmproxy_adapter import MitmproxyRequestAdapter
from .raw_adapter import RawRequestAdapter
from .stoobly_adapter import StooblyRequestAdapter

class PythonRequestAdapterFactory():

  def __init__(self, request: requests.Request):
    self.__request = request

  def mitmproxy_request(self, http_version: str = 'HTTP/1.1'):
    return MitmproxyRequestAdapter(http_version, self.__request).adapt()

  def raw_request(self, http_version: str = 'HTTP/1.1'):
    return RawRequestAdapter(http_version, self.__request).adapt()

  def stoobly_request(self):
    return StooblyRequestAdapter(self.__request).adapt()


import urllib

from ..logger import Logger
from .stoobly_api import StooblyApi

class ReportsResource(StooblyApi):

  def create(self, project_id: str, params = {}):
    url = f"{self.service_url}{self.REPORTS_ENDPOINT}"

    body = {
        'project_id': project_id,
        **params,
    }

    return self.post(url, headers=self.default_headers, json=body)

  def index(self, project_id: str, query_params = {}):
    url = f"{self.service_url}{self.REPORTS_ENDPOINT}"

    return self.get(url, headers=self.default_headers, params={
        'project_id': project_id,
        **query_params,
    })


import hashlib
import pdb
import requests

from mitmproxy.coretypes.multidict import MultiDict
from typing import List, Dict, TypedDict, Union

from stoobly_agent.lib.logger import Logger, bcolors
from stoobly_agent.lib.utils.python_to_ruby_type import type_map

LOG_ID = 'RequestHasher'

class IgnoredParam(TypedDict):
  inferred_type: str
  query: str

class RequestHasher():

  _instance = None

  def __init__(self):
    if self._instance:
        raise RuntimeError('Call instance() instead')
    else:
        self.type_map = type_map

  @classmethod
  def instance(cls):
      if cls._instance is None:
        cls._instance = cls()

      return cls._instance
  
  def hash_params(self, params: Union[dict, list], ignored_params: Dict[str, IgnoredParam] = {}) -> str:
    if isinstance(params, dict) or isinstance(params, list) or isinstance(params, MultiDict) or isinstance(params, requests.structures.CaseInsensitiveDict):
      return self.__serialize(None, params, None, ignored_params)
    else:
      return ''

  def hash_text(self, text: Union[bytes, str]) -> str:
    if len(text) == 0:
      return ''
    return hashlib.md5(text.encode() if isinstance(text, str) else text).hexdigest()
 
  def __serialize(
    self, key: str, value: Union[dict, list, MultiDict, str], query: str, ignored_params: Dict[str, IgnoredParam] = {}
  ) -> str:
    if isinstance(value, dict) or isinstance(value, MultiDict):
      return self.__serialize_hash(key, value, query, ignored_params)
    elif isinstance(value, list):
      return self.__serialize_array(key, value, query, ignored_params)
    else:
      if self.__ignored(query, ignored_params, value):
        return

      return self.hash_text(self.__serialize_param(query, key, value))

  def __ignored(self, query: str, ignored_params: dict, value: Union[dict, list, str]) -> str:
    param = ignored_params.get(query)

    if not param:
      return False

    # Sometimes there's different types for a param
    if not isinstance(param, list):
      return param['query'] == query 

    for param_type in param:
      if self.__infer_type(value) == param_type['inferred_type'] and param_type['query'] == query:
        return True

  def __serialize_array(
    self, key: str, value: Union[dict, list, str], query: str, ignored_params: Dict[str, IgnoredParam] = {}
  ) -> str:
    serialized_params = []

    for v in value:
      param_hash = self.__serialize(key, v, '[*]' if not query else f"{query}[*]", ignored_params)

      if not param_hash:
        continue

      serialized_params.append(param_hash)

    return self.__hash_serialized_params(serialized_params)

  def __serialize_hash(
    self, key: str, value: Union[dict, list, str], query: str, ignored_params: Dict[str, IgnoredParam] = {}
  ) -> str:
    serialized_params = []

    for k, v in value.items():
      param_hash = self.__serialize(k, v, k if not query else f"{query}.{k}", ignored_params)

      if not param_hash:
        continue

      serialized_params.append(param_hash)

    return self.__hash_serialized_params(serialized_params)

  def __hash_serialized_params(self, serialized_params: List[str]):
    if not serialized_params or len(serialized_params) == 0:
      return '' 
    serialized_params.sort()
    return self.hash_text('.'.join(serialized_params))

  def __infer_type(self, val):
    return self.type_map.get(str(val.__class__))

  def __serialize_param(self, query, key: str, val):
    Logger.instance(LOG_ID).debug(f"{bcolors.OKBLUE}Serializing{bcolors.ENDC} {query} -> {val}")

    if isinstance(val, bool):
      # Ruby boolean are lower case
      val = str(val).lower()
    elif val == None:
      # Ruby str(nil) translates to ''
      val = ''

    return f"{str(key)}.{str(val)}"



import requests
import urllib
import pdb

from ..logger import Logger
from .stoobly_api import StooblyApi

LOG_ID = 'UsersResource'

class UsersResource(StooblyApi):
  USERS_ENDPOINT = 'users'

  def profile(self, query_params = {}) -> requests.Response:
    url = f"{self.service_url}/{self.USERS_ENDPOINT}/profile"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)


import requests
import pdb

from stoobly_orator.migrations import Migration

from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.app.proxy.mock.request_hasher import RequestHasher

class AlignRequests(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('requests') as table:
            table.string('response_hash').default('')
            table.string('response_headers_hash').default('')

        for request in Request.all():
            response = request.response
            if not response:
                continue

            _response: requests.Response = RawHttpResponseAdapter(response.raw).to_response()
            response_hash = RequestHasher.instance().hash_text(_response.content)
            response_headers_hash = RequestHasher.instance().hash_params(_response.headers)

            request.update(response_hash=response_hash, response_headers_hash=response_headers_hash)

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('requests') as table:
            table.drop_column('response_hash', 'response_headers_hash')





import os
import pdb
import yaml

from runpy import run_path
from typing import List, Union

from mitmproxy.http import Request as MitmproxyRequest

from stoobly_agent.app.settings.constants import firewall_action, intercept_mode
from stoobly_agent.app.settings.rewrite_rule import RewriteRule
from stoobly_agent.app.settings.firewall_rule import FirewallRule
from stoobly_agent.app.settings import Settings
from stoobly_agent.app.settings.types import IgnoreRule, MatchRule
from stoobly_agent.config.constants import custom_headers, env_vars, mode, request_origin, test_filter
from stoobly_agent.lib.api.keys.project_key import InvalidProjectKey, ProjectKey
from stoobly_agent.lib.logger import Logger

class InterceptSettings:

  def __init__(self, settings: Settings, request: MitmproxyRequest = None):
    self.__settings = settings
    self.__request = request
    self.__headers: MitmproxyRequest.headers = request.headers if request else None
    self.__for_response = False

    parsed_project_key = self.parsed_project_key
    project_id = parsed_project_key.id if parsed_project_key else None

    # If no valid project key is provided, use default settings,
    # Otherwise, set settings for the project
    self.__data_rules = self.__settings.proxy.data.data_rules(project_id)
    self.__rewrite_rules = self.__settings.proxy.rewrite.rewrite_rules(project_id)
    self.__firewall_rules = self.__settings.proxy.firewall.firewall_rules(project_id)
    self.__intercept_settings = self.__settings.proxy.intercept 
    self.__match_rules = self.__settings.proxy.match.match_rules(project_id)

    self.__lifecycle_hooks = None
    self.__initialize_lifecycle_hooks()

    self.__response_fixtures = None
    self.__initialize_response_fixtures()

    self._mock_rewrite_rules = None
    self._record_rewrite_rules = None
    self._replay_rewrite_rules = None
    self._test_rewrite_rules = None

  @property
  def settings(self):
    return self.__settings

  @property
  def alias_resolve_strategy(self):
    if self.__headers and custom_headers.ALIAS_RESOLVE_STRATEGY in self.__headers:
      return self.__headers[custom_headers.ALIAS_RESOLVE_STRATEGY]

  @property
  def active(self):
    if self.__intercept_settings.active:
      return True

    if not self.__headers:
      return False
    
    return custom_headers.PROXY_MODE in self.__headers

  @property
  def lifecycle_hooks_path(self):
    if self.__headers and custom_headers.LIFECYCLE_HOOKS_PATH in self.__headers:
      return self.__headers[custom_headers.LIFECYCLE_HOOKS_PATH]

    if os.environ.get(env_vars.AGENT_LIFECYCLE_HOOKS_PATH):
      return os.environ[env_vars.AGENT_LIFECYCLE_HOOKS_PATH] 

  @property
  def lifecycle_hooks(self):
    return self.__lifecycle_hooks or {}

  @property
  def mode(self):
    if self.__headers:
      if self.__for_response and custom_headers.RESPONSE_PROXY_MODE in self.__headers:
        return self.__headers[custom_headers.RESPONSE_PROXY_MODE]

      access_control_header = self.__headers.get('Access-Control-Request-Headers')
      do_proxy_header = custom_headers.DO_PROXY

      if access_control_header and do_proxy_header.lower() in access_control_header:
          return mode.NONE

      if do_proxy_header in self.__headers:
          return mode.NONE

      if custom_headers.PROXY_MODE in self.__headers:
          return self.__headers[custom_headers.PROXY_MODE]

    return self.__intercept_settings.mode

  @property
  def project_key(self):
    if self.__headers and custom_headers.PROJECT_KEY in self.__headers:
        return self.__headers[custom_headers.PROJECT_KEY]

    return self.__settings.proxy.intercept.project_key

  @property
  def parsed_project_key(self):
    try: 
      return ProjectKey(self.project_key)
    except InvalidProjectKey:
      pass

  @property
  def public_directory_path(self):
    if self.__headers and custom_headers.PUBLIC_DIRECTORY_PATH in self.__headers:
      return self.__headers[custom_headers.PUBLIC_DIRECTORY_PATH]

    if os.environ.get(env_vars.AGENT_PUBLIC_DIRECTORY_PATH):
      return os.environ[env_vars.AGENT_PUBLIC_DIRECTORY_PATH] 

  @property
  def remote_project_key(self):
    # When not local project, don't return set remote project_key
    project_key = self.parsed_project_key
    if project_key and not project_key.is_local:
      return 

    if self.__headers and custom_headers.REMOTE_PROJECT_KEY in self.__headers:
      return self.__headers[custom_headers.REMOTE_PROJECT_KEY]

    return self.__settings.remote.project_key

  @property
  def response_fixtures_path(self):
    if self.__headers and custom_headers.RESPONSE_FIXTURES_PATH in self.__headers:
      return self.__headers[custom_headers.RESPONSE_FIXTURES_PATH]

    if os.environ.get(env_vars.AGENT_RESPONSE_FIXTURES_PATH):
      return os.environ[env_vars.AGENT_RESPONSE_FIXTURES_PATH] 

  @property
  def response_fixtures(self):
    return self.__response_fixtures or {}

  @property
  def parsed_remote_project_key(self):
    try: 
      return ProjectKey(self.remote_project_key)
    except InvalidProjectKey:
      pass
    
  @property
  def scenario_key(self):
    if self.__headers and custom_headers.SCENARIO_KEY in self.__headers:
        return self.__headers[custom_headers.SCENARIO_KEY]

    return self.__data_rules.scenario_key

  @scenario_key.setter
  def scenario_key(self, v):
    self.__data_rules.scenario_key = v

  @property
  def report_key(self) -> Union[str, None]:
    if self.__headers and custom_headers.REPORT_KEY in self.__headers:
      return self.__headers[custom_headers.REPORT_KEY]

  @property
  def policy(self):
    return self.__policy(self.mode)

  @property
  def response_policy(self):
    if self.__headers and custom_headers.RESPONSE_PROXY_MODE in self.__headers:
      mode = self.__headers[custom_headers.RESPONSE_PROXY_MODE]
      return self.__policy(mode)

    return self.policy

  @property
  def exclude_rules(self) -> List[FirewallRule]:
    _mode = self.mode
    return list(filter(lambda rule: _mode in rule.modes and rule.action == firewall_action.EXCLUDE, self.__firewall_rules))

  @property
  def include_rules(self) -> List[FirewallRule]:
    _mode = self.mode
    return list(filter(lambda rule: _mode in rule.modes and rule.action == firewall_action.INCLUDE, self.__firewall_rules))

  @property
  def match_rules(self) -> List[MatchRule]:
    _mode = self.mode
    return list(filter(lambda rule: _mode in rule.modes, self.__match_rules))

  # TODO: explore if should support specifying components to ignore
  @property
  def ignore_rules(self) -> List[IgnoreRule]:
    return []

  @property
  def rewrite_rules(self) -> List[RewriteRule]:
    _mode = self.mode
    return self.__select_rewrite_rules(_mode)

  @property
  def record_rewrite_rules(self) -> List[RewriteRule]:
    if not self._record_rewrite_rules:
      self._record_rewrite_rules = self.__select_rewrite_rules(mode.RECORD)
    return self._record_rewrite_rules

  @property
  def mock_rewrite_rules(self) -> List[RewriteRule]:
    if not self._mock_rewrite_rules:
      self._mock_rewrite_rules = self.__select_rewrite_rules(mode.MOCK)
    return self._mock_rewrite_rules

  @property
  def replay_rewrite_rules(self) -> List[RewriteRule]:
    if not self._replay_rewrite_rules:
      self._replay_rewrite_rules = self.__select_rewrite_rules(mode.REPLAY)
    return self._replay_rewrite_rules

  @property
  def test_rewrite_rules(self) -> List[RewriteRule]:
    if not self._test_rewrite_rules:
      self._test_rewrite_rules = self.__select_rewrite_rules(mode.TEST)
    return self._test_rewrite_rules

  @property
  def upstream_url(self):
    if self.__headers and custom_headers.SERVICE_URL in self.__headers:
      return self.__headers[custom_headers.SERVICE_URL]

    settings_upstream_url = self.__intercept_settings.upstream_url
    if settings_upstream_url and len(settings_upstream_url) > 0:
      return self.__intercept_settings.upstream_url

    if self.__request:
      return f"{self.__request.scheme}://{self.__request.host}:{self.__request.port}"

  @property
  def test_filter(self):
    if self.__headers and custom_headers.TEST_FILTER in self.__headers:
      return self.__headers[custom_headers.TEST_FILTER]

    return test_filter.ALL

  @property
  def test_save_results(self):
    if self.__headers and custom_headers.TEST_SAVE_RESULTS in self.__headers:
      return not not int(self.__headers[custom_headers.TEST_SAVE_RESULTS])

    return False

  @property
  def test_skip(self):
    if self.__headers and custom_headers.TEST_SKIP in self.__headers:
      return True
    
    return False
  
  @property
  def test_strategy(self):
    if self.__headers and custom_headers.TEST_STRATEGY in self.__headers:
      return self.__headers[custom_headers.TEST_STRATEGY]

    return self.__data_rules.test_strategy

  @property
  def request_origin(self):
    if self.__headers and custom_headers.REQUEST_ORIGIN in self.__headers:
      return self.__headers[custom_headers.REQUEST_ORIGIN]

    return request_origin.WEB

  def for_response(self):
    self.__for_response = True

  def __select_rewrite_rules(self, mode = None):
    rules = []

    # Filter only parameters matching active intercept mode
    for rewrite_rule in self.__rewrite_rules:
      # If url rule applies, then update .url_rules with url_rule
      url_rules = self.__select_url_rules(rewrite_rule)

      # If parameters rules apply, then update .parameter_rules with applicable parameter_rules
      parameter_rules = self.__select_parameter_rules(rewrite_rule, mode)

      if len(url_rules) > 0 or len(parameter_rules) > 0:
        # Build a new RewriteRule object contain only parameter rules matching intercept mode
        rewrite_rule = RewriteRule(rewrite_rule.to_dict())
        rewrite_rule.url_rules = url_rules
        rewrite_rule.parameter_rules = parameter_rules
        rules.append(rewrite_rule)

    return rules

  def __select_parameter_rules(self, rewrite_rule: RewriteRule, mode = None):
    mode = mode or self.mode
    return list(filter(
      lambda parameter: mode in parameter.modes and parameter.name, 
      rewrite_rule.parameter_rules or []
    ))

  def __select_url_rules(self, rewrite_rule: RewriteRule, mode = None):
    mode = mode or self.mode
    return list(filter(
      lambda url: mode in url.modes,
      rewrite_rule.url_rules or []
    ))

  def __initialize_lifecycle_hooks(self):
    script_path = self.lifecycle_hooks_path

    if not script_path:
        return

    if not os.path.exists(script_path):
        return Logger.instance().error(f"Lifecycle hooks script {script_path} does not exist")

    try:
        self.__lifecycle_hooks = run_path(script_path)
    except Exception as e:
        return Logger.instance().error(e)

  def __initialize_response_fixtures(self):
    fixtures_path = self.response_fixtures_path

    if not fixtures_path:
        return

    if not os.path.exists(fixtures_path):
        return Logger.instance().error(f"Response fixtures {fixtures_path} does not exist")

    with open(fixtures_path, 'r') as stream:
        try:
            self.__response_fixtures = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            Logger.instance().error(exc)
        
  def __policy(self, mode):
    if mode == intercept_mode.MOCK:
      if self.__headers and custom_headers.MOCK_POLICY in self.__headers:
        return self.__headers[custom_headers.MOCK_POLICY]

      return self.__data_rules.mock_policy
    elif mode == intercept_mode.RECORD:
      if self.__headers and custom_headers.RECORD_POLICY in self.__headers:
        return self.__headers[custom_headers.RECORD_POLICY]

      return self.__data_rules.record_policy
    elif mode == intercept_mode.TEST:
      return self.__data_rules.test_policy
    elif mode == intercept_mode.REPLAY:
      return self.__data_rules.replay_policy


import pdb
import requests

from stoobly_agent.app.proxy.mitmproxy.request_facade import MitmproxyRequestFacade
from stoobly_agent.app.proxy.record.proxy_request import ProxyRequest
from stoobly_agent.app.proxy.record.request_string import RequestString

from .mitmproxy_adapter import MitmproxyRequestAdapter

class RawRequestAdapter():

  def __init__(self, http_version: str, request: requests.Request):
    self.__http_version = http_version
    self.__request = request

  def adapt(self):
    mitmproxy_request = MitmproxyRequestAdapter(self.__http_version, self.__request).adapt()
    adapted_request = MitmproxyRequestFacade(mitmproxy_request)
    proxy_request = ProxyRequest(adapted_request)
    request_string = RequestString(proxy_request) 
    return request_string.get()


import uuid

from .uuid_key import UuidKey

class InvalidScenarioKey(Exception):
  pass

class ScenarioKey(UuidKey):

  def __init__(self, key: str):
    super().__init__(key)

    self.__raw = key

    if not self.id:
      raise InvalidScenarioKey('Missing id')

    if not self.project_id:
      raise InvalidScenarioKey('Missing project_id')

  @property
  def project_id(self) -> str:
    return self.get('p')

  @property
  def raw(self):
    return self.__raw

  @staticmethod
  def encode(project_id: str, request_id: str):
    return UuidKey.encode({
      'p': project_id,
      'i': request_id.replace('-', ''),
    })


from stoobly_orator.migrations import Migration

class AddFilterColumnsToRequests(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('requests') as table:
            table.boolean('is_deleted').default(False)
            table.boolean('starred').default(False)

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('requests') as table:
            table.drop_column('is_deleted', 'starred')



from .json_print import json_print
from .tabulate_print_service import tabulate_print
from ..types.print_options import TabulatePrintOptions

JSON_FORMAT = 'json'
SIMPLE_FORMAT = 'simple'
FORMATS = [JSON_FORMAT, SIMPLE_FORMAT]

def select_print_options(kwargs: TabulatePrintOptions):
    print_options = {
        'format': kwargs['format'],
        'select': kwargs['select'],
        'without_headers': kwargs['without_headers']
    }

    del kwargs['format']
    del kwargs['select']
    del kwargs['without_headers']

    return print_options

def print_projects(projects, **kwargs: TabulatePrintOptions):
    filter = ['created_at', 'is_deleted', 'organization_id', 'project_id', 'starred', 'updated_at']
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(projects, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
            projects, 
            filter=filter,
            headers=not kwargs.get('without_headers'),
            select=kwargs.get('select') or []
        )

def print_reports(reports, **kwargs: TabulatePrintOptions):
    filter = ['created_at', 'priority', 'user_id', 'starred', 'updated_at']
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(reports, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
            reports, 
            filter=filter,
            headers=not kwargs.get('without_headers'),
            select=kwargs.get('select') or []
        )

def print_requests(requests, **kwargs: TabulatePrintOptions):
    filter = [
        'body_params_hash', 'body_text_hash', 'components' , 'created_at', 'endpoint', 'endpoint_id', 'http_version', 'is_deleted', 'position', 'project_id', 'pushed_at', 'query_params_hash', 'scenario_id', 'scheme', 'starred', 'uuid', 'updated_at', 'url'
    ]
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(requests, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
            requests, 
            filter=filter,
            headers=not kwargs.get('without_headers'),
            select=kwargs.get('select') or []
        )

def print_scenarios(scenarios, **kwargs: TabulatePrintOptions):
    filter = ['created_at', 'is_deleted', 'priority', 'project_id', 'starred', 'uuid', 'updated_at']
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(scenarios, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
            scenarios, 
            filter=filter,
            headers=not kwargs.get('without_headers'),
            select=kwargs.get('select') or []
        )

def print_services(services, **kwargs: TabulatePrintOptions):
    filter = []
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(services, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
            services, 
            filter=filter,
            headers=not kwargs.get('without_headers'),
            select=kwargs.get('select') or []
        )

def print_snapshots(snapshots, **kwargs: TabulatePrintOptions):
    filter = ['resource_uuid']
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(snapshots, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
            snapshots,
            filter=filter,
            headers=not kwargs.get('without_headers'),
            select=kwargs.get('select') or []
        )

def print_tests(tests, **kwargs: TabulatePrintOptions):
    filter = ['created_at', 'id', 'log', 'position', 'project_id', 'report_id', 'scenario_id', 'starred', 'updated_at']
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(tests, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
            tests, 
            filter=filter,
            headers=not kwargs.get('without_headers'),
            select=kwargs.get('select') or []
        )

def print_traces(traces, **kwargs: TabulatePrintOptions):
    filter = kwargs.get('filter') or []
    format = kwargs.get('format')

    if format == JSON_FORMAT:
        json_print(traces, **{
            'filter': filter,
            **kwargs
        })
    else:
        tabulate_print(
        traces, 
        filter=filter,
        headers=not kwargs.get('without_headers'),
        select=kwargs.get('select') or []
        )


import pdb

from stoobly_agent.lib.api.keys.project_key import ProjectKey
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey

class ParamBuilder():
  def __init__(self, initial_params = {}):
    self.__params = initial_params

  def with_resource_scoping(self, _project_key: str, _scenario_key: str = None):
    # Try to see if a scenario is set, otherwise use project
    if _scenario_key and len(_scenario_key) > 0:
        scenario_key = ScenarioKey(_scenario_key)
        self.__params['scenario_id'] = scenario_key.id
        self.__params['project_id'] = scenario_key.project_id
    elif _project_key and len(_project_key) > 0:
        project_key = ProjectKey(_project_key)
        self.__params['project_id'] = project_key.id

    return self

  def with_params(self, params):
    self.__params = { **self.__params, **params }
    return self

  def with_param(self, key, val):
    self.__params[key] = val
    return self

  def build(self) -> dict:
    return self.__params


import base64
import json
import os
import pdb

from stoobly_agent.config.constants import custom_headers, env_vars
from stoobly_agent.lib.api.keys import InvalidReportKey, InvalidScenarioKey, ProjectKey, ReportKey, ScenarioKey

from .api import Api

class StooblyApi(Api):
    REPORTS_ENDPOINT = '/reports'
    REQUESTS_ENDPOINT = '/requests'
    TESTS_ENDPOINT = '/tests'

    def __init__(self, service_url: str, api_key: str):
        self.service_url = service_url
        self.api_key = api_key

    @property
    def default_headers(self):
        headers = {
            'X-API-KEY': self.api_key,
        }

        if not os.environ.get(env_vars.AGENT_SELF_INTERCEPT_ENABLED):
            headers[custom_headers.DO_PROXY] = '1'

        return headers

    # Request

    def from_project_key(self, project_key, handler):
        key = ProjectKey(project_key)
        return handler(key.id)

    def from_scenario_key(self, scenario_key, handler):
        key = ScenarioKey(scenario_key)
        return handler(key.project_id, {
            'scenario_id': key.id,
        })

    def with_report_key(self, report_key, params):
        try:
            key = ReportKey(report_key)
            params['report_id'] = key.id
        except InvalidReportKey:
            pass

        return self

    def with_scenario_key(self, scenario_key, params):
        try:
            key = ScenarioKey(scenario_key)
            params['scenario_id'] = key.id
        except InvalidScenarioKey:
            pass

        return self


from .endpoints import EndpointsIndexQueryParams
from .headers import Header, HeaderShowResponse
from .pagination_query_params import PaginationQueryParams
from .projects import ProjectCreateParams, ProjectDetails, ProjectShowResponse, ProjectsIndexQueryParams, ProjectsIndexResponse
from .query_params import QueryParam, QueryParamShowResponse
from .reports import ReportShowResponse
from .requests import RequestCreateParams, RequestResponseShowQueryParams, RequestShowQueryParams, RequestShowResponse, RequestsIndexQueryParams, RequestsIndexResponse
from .response_headers import ResponseHeaderShowResponse
from .responses import ResponseShowResponse
from .scenarios import ScenarioShowResponse, ScenariosIndexQueryParams, ScenariosIndexResponse
from .tests import TestCreateParams, TestShowResponse, TestsIndexQueryParams, TestsIndexResponse
from .traces import TraceShowResponse
from .users import UserProfileResponse



import pdb
import requests

from urllib.parse import parse_qs, urlparse

class StooblyRequestAdapter():

  def __init__(self, request: requests.Request):
    self.__request = request

  def adapt(self):
    parsed_url = urlparse(self.__request.url)

    return {
      'body': self.__request.data,
      'headers': self.adapt_headers(),
      'method': self.__request.method,
      'path': parsed_url.path,
      'query_params': self.adapt_query_params(parsed_url.query),
      'url': self.__request.url,
    }

  def adapt_query_params(self, query = None):
    if not query:
      parsed_url = urlparse(self.__request.url)
      query = parsed_url.query

    _query_params = parse_qs(query)

    query_params = []
    for k, v in _query_params.items():
      for param in v:
        query_params.append({
          'name': k,
          'value': param,
        })

    return query_params

  def adapt_headers(self):
    headers = []

    for k, v in self.__request.headers.items():
      headers.append({
        'name': k,
        'value': v
      })

    return headers


import pdb

from stoobly_agent.app.models.adapters.orm.request.mitmproxy_adapter import MitmproxyRequestAdapter
from stoobly_agent.app.models.types.request import RequestIndexSimilarParams
from stoobly_agent.app.proxy.mitmproxy.request_facade import MitmproxyRequestFacade
from stoobly_agent.app.proxy.mock.hashed_request_decorator import HashedRequestDecorator
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse
from stoobly_agent.lib.orm.request import Request

from .endpoints_apply_context import EndpointsApplyContext

def apply_endpoints(context: EndpointsApplyContext):
  for endpoint in context.endpoints:
    for endpoint_handler in context.endpoint_handlers:
      endpoint_handler(endpoint)

    apply_endpoint(context, endpoint)

def apply_endpoint(context: EndpointsApplyContext, endpoint: EndpointShowResponse):
  ignored_components = endpoint['ignored_components'] if 'ignored_components' in endpoint else []
  requests = endpoint_requests(context, endpoint)

  for request in requests:
    for request_handler in context.request_handlers:
      request_handler(request, endpoint)

    hash_request(request, ignored_components)

def hash_request(request: Request, ignored_components = []):
  mitmproxy_request = MitmproxyRequestAdapter(request).adapt()
  request_facade = MitmproxyRequestFacade(mitmproxy_request)
  hashed_request = HashedRequestDecorator(request_facade)

  hashed_request.with_ignored_components(ignored_components) 

  request.headers_hash = hashed_request.headers_hash()
  request.query_params_hash = hashed_request.query_params_hash()
  request.body_params_hash = hashed_request.body_params_hash()
  request.body_text_hash = hashed_request.body_text_hash()

  request.save()

def endpoint_requests(context: EndpointsApplyContext, endpoint: EndpointShowResponse):
  local_db_request_adapter = context.request_model
  if not local_db_request_adapter:
    return []

  host = endpoint['host']
  if not host:
    host = '%'

  port = endpoint['port']
  if not port:
    port = '%'

  method = endpoint['method']
  pattern = endpoint['match_pattern']

  scenario_id = context.scenario_id
  params = RequestIndexSimilarParams(host=host, port=port, method=method, pattern=pattern, scenario_id=scenario_id)
  return local_db_request_adapter.index_similar(params)


import pdb
import pytest

from urllib.parse import urlparse

from stoobly_agent.app.models.factories.resource.local_db.helpers.request_builder import RequestBuilder
from stoobly_agent.app.models.factories.resource.local_db.request_adapter import (
  LocalDBRequestAdapter,
)
from stoobly_agent.app.models.types.request import RequestIndexSimilarParams
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.test.test_helper import reset

class TestLocalDBRequestAdapter():
  @pytest.fixture(autouse=True, scope='class')
  def settings(self):
    return reset()

  @pytest.fixture(scope='class')
  def local_db_request_adapter(self):
    adapter = LocalDBRequestAdapter(request_orm=Request, response_orm=Response)
    return adapter

  class SimilarRequests():
    @pytest.fixture(scope='function')
    def get_v2_pets(self):
      record = Request.create(
        method='GET',
        scheme='https',
        host="petstore.swagger.io",
        port="443",
        path='/v2/pets',
        headers_hash='c079b917a5091fbcdea198bc96c87796',
        body_text_hash='',
        query_params_hash='90f45a51e497a899f995f5c913b805ba',
        body_params_hash='',
        control='1 oxsmp817uge 1684883577229000000',
        raw='GET https://petstore.swagger.io/v2/pets?limit=1 HTTP/1.1 Content-Length: 0',
        query='limit=1'
      )
      yield record
      record.delete()

    @pytest.fixture(scope='function')
    def get_v2_pet_1(self):
      record = Request.create(
        method='GET',
        scheme='https',
        host="petstore.swagger.io",
        port="443",
        path='/v2/pet/1',
        headers_hash='c079b917a5091fbcdea198bc96c87796',
        body_text_hash='',
        query_params_hash='',
        body_params_hash='',
        control='1 l9jiwridgk 1684883778859000000',
        raw='GET https://petstore.swagger.io/v2/pets/1 HTTP/1.1 Content-Length: 0',
      )
      yield record
      record.delete()

    @pytest.fixture(scope='function')
    def create_put_v3_pet(self):
      record = Request.create(
        method='PUT',
        scheme='http',
        host="swagger.io",
        port="80",
        path='/v3/pet',
        headers_hash='c079b917a5091fbcdea198bc96c87796',
        body_text_hash='',
        query_params_hash='',
        body_params_hash='',
        control='1 gocrqlad4ru 1690266880243000000',
        raw='PUT http://swagger.io/v3/pet HTTP/1.1 Content-Length: 0',
      )
      yield record
      record.delete()

    @pytest.fixture(scope='function')
    def created_scenario(self):
      record = Scenario.create(name="Pets scenario")
      yield record
      record.delete()

    @pytest.mark.openapi
    def test_finds_get_v2_pets(self, local_db_request_adapter: LocalDBRequestAdapter, get_v2_pets: Request, get_v2_pet_1: Request):
      assert get_v2_pets
      assert get_v2_pet_1

      params: RequestIndexSimilarParams = {
        'method': 'GET',
        'host' : "petstore.swagger.io",
        'port': "443",
        'pattern': '/v2/pets'
      }

      similar_requests = local_db_request_adapter.index_similar(params)

      assert len(similar_requests) == 1

      request = similar_requests[0]
      assert request.host == params['host']
      assert request.port == int(params['port'])
      assert request.method == params['method']
      assert request.path == params['pattern']

    @pytest.mark.openapi
    def test_finds_put_v3_pet(self, local_db_request_adapter: LocalDBRequestAdapter, create_put_v3_pet: Request):
      params: RequestIndexSimilarParams = {
        'host': '%',
        'port': '%',
        'method': 'PUT',
        'pattern': '/v3/pet',
        'scenario_id': 0
      }

      similar_requests = local_db_request_adapter.index_similar(params)

      assert len(similar_requests) == 1

      request = similar_requests[0]
      assert request.id == create_put_v3_pet.id
      assert request.method == params['method']
      assert request.path == params['pattern']

    class Scenario():
      @pytest.mark.openapi
      def test_finds_no_requests(
        self, local_db_request_adapter: LocalDBRequestAdapter, get_v2_pets: Request, get_v2_pet_1: Request, created_scenario: Scenario
      ):
        assert get_v2_pets
        assert get_v2_pet_1

        params: RequestIndexSimilarParams = {
          'method': 'GET',
          'host' : "petstore.swagger.io",
          'port': "443",
          'pattern': '/v2/pets',
          'scenario_id': created_scenario.id,
        }

        similar_requests = local_db_request_adapter.index_similar(params)

        assert len(similar_requests) == 0

      @pytest.mark.openapi
      def test_finds_requests(
        self, local_db_request_adapter: LocalDBRequestAdapter, get_v2_pets: Request, get_v2_pet_1: Request, created_scenario: Scenario
      ):
        assert get_v2_pets
        assert get_v2_pet_1

        Request.where('id', get_v2_pets.id).update(scenario_id=created_scenario.id)

        params: RequestIndexSimilarParams = {
          'method': 'GET',
          'host' : "petstore.swagger.io",
          'port': "443",
          'pattern': '/v2/pets',
          'scenario_id': created_scenario.id,
        }

        similar_requests = local_db_request_adapter.index_similar(params)

        assert len(similar_requests) == 1
        assert similar_requests[0].id == get_v2_pets.id

  class TestWhenResponse():
    class TestTiebreakScenarioRequest():
      @pytest.fixture(autouse=True, scope='class')
      def settings(self):
        return reset()

      @pytest.fixture(scope='class')
      def request_method(self):
        return 'POST'

      @pytest.fixture(scope='class')
      def request_url(self):
        return 'https://petstore.swagger.io/v2/pets'

      @pytest.fixture(scope='class')
      def created_scenario(self):
        return Scenario.create(name='test')

      @pytest.fixture(autouse=True, scope='class')
      def created_request_one(
        self, settings: Settings, 
        request_method: str, request_url: str, created_scenario: Scenario,
      ):
        status = RequestBuilder(
          method=request_method,
          request_body='',
          request_headers={},
          response_body='',
          status_code=201,
          url=request_url,
        ).with_settings(settings).build()[1]
        assert status == 200

        request = Request.last()
        request.update(scenario_id=created_scenario.id)

        return request

      @pytest.fixture(autouse=True, scope='class')
      def created_request_two(
        self, settings: Settings, 
        request_method: str, request_url: str, created_scenario: Scenario
      ):
        status = RequestBuilder(
          method=request_method,
          request_body='',
          request_headers={},
          response_body='2',
          status_code=202,
          url=request_url,
        ).with_settings(settings).build()[1]
        assert status == 200

        request = Request.last()
        request.update(scenario_id=created_scenario.id)
        return request

      @pytest.fixture(scope='class')
      def requests(self, created_request_one: Request, created_request_two: Request):
        return [created_request_one, created_request_two]

      def test_it_returns_request_one(self, request_url: str, created_scenario: Scenario, local_db_request_adapter: LocalDBRequestAdapter):
        uri = urlparse(request_url)
        response = local_db_request_adapter.response(
          host=uri.hostname, path=uri.path, scenario_id=created_scenario.id
        )
        assert response.status_code == 201

      def test_it_returns_request_two(self, request_url: str, created_scenario: Scenario, local_db_request_adapter: LocalDBRequestAdapter):
        uri = urlparse(request_url)
        response = local_db_request_adapter.response(
          host=uri.hostname, path=uri.path, scenario_id=created_scenario.id
        )
        assert response.status_code == 202

      def test_it_returns_request_one_again(self, request_url: str, created_scenario: Scenario, local_db_request_adapter: LocalDBRequestAdapter):
        uri = urlparse(request_url)
        response = local_db_request_adapter.response(
          host=uri.hostname, path=uri.path, scenario_id=created_scenario.id
        )
        assert response.status_code == 201



import glob
import os
import pdb
import shutil

from .app import App
from .service_command import ServiceCommand
from stoobly_agent.lib.logger import Logger


LOG_ID = 'ServiceDeleteCommand'

class ServiceDeleteCommand(ServiceCommand):
  def __init__(self, app: App, **kwargs):
    super().__init__(app, **kwargs)

  def delete(self) -> None:
    Logger.instance(LOG_ID).debug(f"Deleting service: {self.service_name}, path: {self.service_path}")

    # Delete service directory
    shutil.rmtree(self.service_path)

    # Delete certs for that hostname if HTTPS
    if self.service_config.scheme == 'https':
      certs_dir_path_escaped = glob.escape(self.app.certs_dir_path)
      hostname = self.service_config.hostname
      pattern = os.path.join(certs_dir_path_escaped, f"{hostname}*")

      cert_paths = glob.glob(pattern)
      for cert_path in cert_paths:
        Logger.instance(LOG_ID).debug(f"Deleting cert_path: {cert_path}")
        os.remove(cert_path)

    return None




import pdb

from ..constants import WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE
from ..docker.workflow.builder import WorkflowBuilder
from .constants import (
  CUSTOM_CONFIGURE, CUSTOM_INIT, CUSTOM_PUBLIC_GITIGNORE, MAINTAINED_CONFIGURE, MOCK_WORKFLOW_CUSTOM_FILES, MOCK_WORKFLOW_MAINTAINED_FILES, RECORD_WORKFLOW_CUSTOM_FILES, RECORD_WORKFLOW_MAINTAINED_FILES, TEST_WORKFLOW_CUSTOM_FILES, TEST_WORKFLOW_MAINTAINED_FILES
)

def custom_files(workflow: str, workflow_builder: WorkflowBuilder):
  files = []
  if workflow == WORKFLOW_MOCK_TYPE:
    files = MOCK_WORKFLOW_CUSTOM_FILES
  elif workflow == WORKFLOW_RECORD_TYPE:
    files = RECORD_WORKFLOW_CUSTOM_FILES
  elif workflow == WORKFLOW_TEST_TYPE:
    files = TEST_WORKFLOW_CUSTOM_FILES
  else:
    if workflow_builder.configure in workflow_builder.services:
      files.append(CUSTOM_CONFIGURE)

    if workflow_builder.init in workflow_builder.services:
      files.append(CUSTOM_INIT)

  # Fixtures are only relevant if the workflow is mock/test and if the service has a hostname
  if not workflow_builder.config.hostname:
    if CUSTOM_PUBLIC_GITIGNORE in files:
      files.remove(CUSTOM_PUBLIC_GITIGNORE)

  return files

def maintained_files(workflow: str, workflow_builder: WorkflowBuilder):
  files = []

  if workflow == WORKFLOW_MOCK_TYPE:
    files = MOCK_WORKFLOW_MAINTAINED_FILES
  elif workflow == WORKFLOW_RECORD_TYPE:
    files = RECORD_WORKFLOW_MAINTAINED_FILES
  elif workflow == WORKFLOW_TEST_TYPE:
    files = TEST_WORKFLOW_MAINTAINED_FILES
  else:
    if workflow_builder.configure in workflow_builder.services:
      files.append(MAINTAINED_CONFIGURE)

  return files


from typing import List, TypedDict, Union

from .pagination_query_params import PaginationQueryParams

class RequestCreateParams(TypedDict):
  importer: str
  project_id: str
  requests: bytes
  scenario_id: str

class RequestResponseShowQueryParams(TypedDict):
  body_params_hash: str
  body_text_hash: str
  headers_hash: str
  host: str
  infer: str
  method: str
  path: str
  port: str
  project_id: str
  query_params_hash: str
  rety: str
  scenario_id: str

class RequestShowQueryParams():
  body: bool
  header: bool
  project_id: str
  query_param: bool
  response: bool

class RequestShowResponse(TypedDict):
  body: str
  components: List[str]
  created_at: str
  endpoint_id: int
  headers: Union[list, None]
  host: str
  http_version: float
  id: str
  is_deleted: bool
  key: str
  latency: int
  method: str
  path: str
  port: str
  query: str
  query_params: Union[list, None]
  scheme: str
  scenario: str
  scenario_id: int
  starred: bool
  status: int
  uuid: str
  url: str

class RequestsIndexQueryParams(PaginationQueryParams):
  filter: str
  project_id: str
  scenario_id: str

class RequestsIndexResponse(TypedDict):
  list: List[RequestShowResponse]
  total: int



import pdb
import re

from mitmproxy.http import Request as MitmproxyRequest
from typing import List

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.settings.firewall_rule import FirewallRule
from stoobly_agent.config.constants import intercept_policy, request_origin
from stoobly_agent.lib.logger import bcolors, Logger

LOG_ID = 'Firewall'

def get_active_mode_policy(request: MitmproxyRequest, intercept_settings: InterceptSettings):
    if intercept_settings.request_origin == request_origin.CLI:
        return intercept_settings.policy 

    if allowed_request(request, intercept_settings):
        return intercept_settings.policy
    else:
        # If the request path does not match accepted paths, do not intercept
        return intercept_policy.NONE

def allowed_request(request: MitmproxyRequest, intercept_settings: InterceptSettings) -> bool:
    # If an exclude rule(s) exists, then only requests not matching these pattern(s) are allowed
    exclude_rules = intercept_settings.exclude_rules
    if __request_excluded(request, exclude_rules):
        return False

    # If an include rule(s) exists, then only requests matching these pattern(s) are allowed
    include_rules = intercept_settings.include_rules
    if not __request_included(request, include_rules):
        return False

    # If there are no exclude or include patterns, request is allowed
    return True

def __request_excluded(request: MitmproxyRequest, exclude_rules: List[FirewallRule]):
    if exclude_rules:
        method = request.method.upper()
        rules = list(filter(lambda rule: method in rule.methods, exclude_rules))
        patterns = list(map(lambda rule: rule.pattern, rules))
        if __exclude(request, patterns):
            Logger.instance(LOG_ID).info(f"{bcolors.OKBLUE}Excluding{bcolors.ENDC} {request.method} {request.url}")
            return True
    
    return False

def __request_included(request: MitmproxyRequest, include_rules: List[FirewallRule]):
    if not include_rules:
        return True

    method = request.method.upper()
    rules = list(filter(lambda rule: method in rule.methods, include_rules))

    # If there are include rules, but none that match the request's method,
    # then we know that none of the include rules will match the request
    if len(include_rules) > 0 and len(rules) == 0:
        Logger.instance(LOG_ID).info(f"{bcolors.OKBLUE}Not Including{bcolors.ENDC} {request.method} {request.url}")
        return False

    patterns = list(map(lambda rule: rule.pattern, rules))
    if not __include(request, patterns):
        Logger.instance(LOG_ID).info(f"{bcolors.OKBLUE}Not Including{bcolors.ENDC} {request.method} {request.url}")
        return False

    return True

###
#
# @param patterns [Array<string>]
#
def __include(request: MitmproxyRequest, patterns: List[str]) -> bool:
    if not patterns:
        return True

    if len(patterns) == 0:
        return True

    for pattern in patterns:
        try:
            if re.match(pattern, request.url):
                return True
        except re.error as e:
            Logger.instance(LOG_ID).error(f"RegExp error '{e}' for {pattern}")
            return False

    return False

def __exclude(request: MitmproxyRequest, patterns: List[str]) -> bool:
    if not patterns:
        return False

    for pattern in patterns:
        try:
            if re.match(pattern, request.url):
                return True
        except re.error as e:
            Logger.instance(LOG_ID).error(f"RegExp error '{e}' for {pattern}")
            return True

    return False



import pdb

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from typing import List

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.settings.rewrite_rule import RewriteRule

from ..mitmproxy.request_facade import MitmproxyRequestFacade
from ..mitmproxy.response_facade import MitmproxyResponseFacade
from ..utils.rewrite import rewrite_request_response
from .joined_request import JoinedRequest
from .proxy_request import ProxyRequest

def join_request(
    adapted_request: MitmproxyRequestFacade, adapted_response: MitmproxyResponseFacade, intercept_settings: InterceptSettings
) -> JoinedRequest:
    # Decorate request with service_url
    upstream_url = intercept_settings.upstream_url
    proxy_request = ProxyRequest(adapted_request, upstream_url)

    # Create JoinedRequest
    return JoinedRequest(proxy_request).with_response(adapted_response)

def join_request_from_flow(
    flow: MitmproxyHTTPFlow, intercept_settings: InterceptSettings
) -> JoinedRequest:
    request = MitmproxyRequestFacade(flow.request)
    response = MitmproxyResponseFacade(flow.response)

    return join_request(request, response, intercept_settings)


from stoobly_agent.app.cli.scaffold.service_config import ServiceConfig

from ...constants import WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE
from .dns_decorator import DnsDecorator
from .mock_decorator import MockDecorator
from .reverse_proxy_decorator import ReverseProxyDecorator

def get_workflow_decorators(workflow: str, service_config: ServiceConfig):
  workflow_decorators = []

  if workflow == WORKFLOW_RECORD_TYPE:
    if service_config.hostname:
      workflow_decorators.append(ReverseProxyDecorator)
      workflow_decorators.append(DnsDecorator)
  elif workflow == WORKFLOW_MOCK_TYPE or workflow == WORKFLOW_TEST_TYPE:
     if service_config.hostname:
      workflow_decorators.append(ReverseProxyDecorator if service_config.detached else MockDecorator)    
      workflow_decorators.append(DnsDecorator)
  else:
    if service_config.hostname:
      workflow_decorators.append(ReverseProxyDecorator if service_config.detached else MockDecorator) 

  return workflow_decorators


import json
import pdb
import pytest

from click.testing import CliRunner, Result

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.models.adapters.raw_joined.response import RawJoinedResponseAdapterFactory
from stoobly_agent.app.models.adapters.orm.joined_request_string_adapter import JoinedRequestStringAdapter
from stoobly_agent.cli import record
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

class TestJoindRequestStringAdapter():
  @pytest.fixture(scope='class', autouse=True)
  def settings(self):
    return reset()

  class TestWhenEncoded():
    @pytest.fixture(scope='class')
    def record_result(self, runner: CliRunner) -> Result:
      url = DETERMINISTIC_GET_REQUEST_URL
      record_result = runner.invoke(record, ['--header', 'Accept-Encoding: gzip', url])
      return record_result

    @pytest.fixture(scope='class')
    def created_request(self, record_result: Result):
      assert record_result.exit_code == 0
      return Request.last()

    def test_idempotent(self, created_request: Request):
      adapter = JoinedRequestStringAdapter(created_request)
      assert adapter.adapt() == adapter.adapt()

    def test_it_decodes(self, created_request: Request):
      adapter = JoinedRequestStringAdapter(created_request)

      non_decoded_response = RawJoinedResponseAdapterFactory(adapter.adapt()).python_response()
      adapter.decode_response()

      decoded_response = RawJoinedResponseAdapterFactory(adapter.adapt()).python_response()

      assert len(non_decoded_response.content) == len(decoded_response.content)

  class TestWhenNotEncoded():
    @pytest.fixture(scope='class')
    def record_result(self, runner: CliRunner) -> Result:
      url = DETERMINISTIC_GET_REQUEST_URL
      record_result = runner.invoke(record, [url])
      return record_result

    @pytest.fixture(scope='class')
    def created_request(self, record_result: Result):
      assert record_result.exit_code == 0
      return Request.last()

    def test_idempotent(self, created_request: Request):
      adapter = JoinedRequestStringAdapter(created_request)
      assert adapter.adapt() == adapter.adapt()

    def test_it_decodes(self, created_request: Request):
      adapter = JoinedRequestStringAdapter(created_request)

      non_decoded_response = RawJoinedResponseAdapterFactory(adapter.adapt()).python_response()
      adapter.decode_response()

      decoded_response = RawJoinedResponseAdapterFactory(adapter.adapt()).python_response()

      assert len(non_decoded_response.content) == len(decoded_response.content)


import abc

from mitmproxy.http import Request
from typing import Union

from stoobly_agent.app.proxy.test.helpers.endpoint_facade import EndpointFacade
from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.lib.api.interfaces.endpoints import EndpointShowResponse
from stoobly_agent.lib.orm.trace import Trace

from .context_response import TestContextResponse

FuzzyContent = Union[dict, list, str]

class TestContextABC(abc.ABC):

  @abc.abstractmethod
  def with_endpoints_resource(self, resource: EndpointShowResponse):
    pass

  @property
  @abc.abstractmethod
  def cached_rewritten_expected_response_content(self) -> FuzzyContent:
    pass

  @property
  @abc.abstractmethod
  def decoded_response_content(self) -> FuzzyContent:
    pass

  @property
  @abc.abstractmethod
  def decoded_expected_response_content(self) -> FuzzyContent:
    pass

  @property
  @abc.abstractmethod
  def endpoint(self) -> EndpointFacade:
    pass

  @property
  @abc.abstractmethod
  def end_time(self):
    pass

  @property
  @abc.abstractmethod
  def expected_latency(self) -> int:
    pass

  @property
  @abc.abstractmethod
  def expected_response(self) -> TestContextResponse:
    pass

  @property
  @abc.abstractmethod
  def expected_status_code(self) -> int:
    pass

  @property
  @abc.abstractmethod
  def filter(self):
    pass

  @property
  @abc.abstractmethod
  def flow(self):
    pass

  @property
  @abc.abstractmethod
  def intercept_settings(self) -> InterceptSettings:
    pass

  @property
  @abc.abstractmethod
  def lifecycle_hooks(self):
    pass

  @property
  @abc.abstractmethod
  def lifecycle_hooks_path(self):
    pass

  @property
  @abc.abstractmethod
  def log(self):
    pass

  @log.setter
  @abc.abstractmethod
  def log(self, v: str):
    pass

  @property
  @abc.abstractmethod
  def mock_context(self):
    pass

  @property
  @abc.abstractmethod
  def mock_request_id(self) -> Union[str, None]:
    pass

  @property
  @abc.abstractmethod
  def mock_request_endpoint_id(self) -> Union[str, None]:
    pass

  @property
  @abc.abstractmethod
  def passed(self):
    pass

  @passed.setter
  @abc.abstractmethod
  def passed(self, v):
    pass

  @property
  @abc.abstractmethod
  def public_directory_path(self):
    pass

  @property
  @abc.abstractmethod
  def replay_context(self):
    pass

  @property
  @abc.abstractmethod
  def request(self) -> Request:
    pass

  @property
  @abc.abstractmethod
  def request_key(self) -> str:
    pass

  @property
  @abc.abstractmethod
  def request_headers(self):
    pass

  @property
  @abc.abstractmethod
  def response(self) -> TestContextResponse:
    pass

  @property
  @abc.abstractmethod
  def response_fixtures(self):
    pass

  @property
  @abc.abstractmethod
  def response_fixtures_path(self):
    pass

  @property
  @abc.abstractmethod
  def response_param_names(self) -> RequestComponentNamesFacade: 
    pass

  @property
  @abc.abstractmethod
  def rewritten_expected_response_content(self):
    pass

  @property
  @abc.abstractmethod
  def skipped(self):
    pass

  @skipped.setter
  @abc.abstractmethod
  def skipped(self, v: bool):
    pass

  @property
  @abc.abstractmethod
  def start_time(self):
    pass

  @property
  @abc.abstractmethod
  def save(self):
    pass

  @property
  @abc.abstractmethod
  def strategy(self):
    pass

  @property
  @abc.abstractmethod
  def trace(self) -> Union[Trace, None]:
    pass



import pdb

from requests import Response
from typing import Callable, Dict, List, Union

from stoobly_agent.app.proxy.replay.alias_resolver import AliasResolver
from stoobly_agent.app.cli.helpers.tabulate_print_service import tabulate_print
from stoobly_agent.app.models.schemas.request import Request
from stoobly_agent.app.proxy.mock.search_endpoint import search_endpoint
from stoobly_agent.app.proxy.replay.body_parser_service import decode_response, is_traversable
from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.proxy.replay.rewrite_params_service import build_id_to_alias_map, rewrite_params
from stoobly_agent.config.constants import alias_resolve_strategy, custom_headers
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.interfaces.endpoints import Alias, EndpointShowResponse, RequestComponentName, ResponseParamName
from stoobly_agent.lib.api.keys import ProjectKey
from stoobly_agent.lib.logger import bcolors, Logger
from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias
from stoobly_agent.lib.orm.trace_request import TraceRequest
from stoobly_agent.lib.utils import jmespath

AliasMap = Dict[str, RequestComponentName]
LOG_ID = 'Trace'

class TraceContext:

  def __init__(self, endpoints_resource: EndpointsResource, trace = None):
    self.__endpoints_resource = endpoints_resource
    self.__trace = trace or Trace.create()

    self.__alias_resolver = AliasResolver(self.__trace, alias_resolve_strategy.NONE)
    self.__requests: List[Request] = []
    self.__remote_project_key: ProjectKey = None

    Logger.instance(LOG_ID).debug(f"Using trace {self.__trace.id}")

  @property
  def alias_resolve_strategy(self):
    return self.__alias_resolver.strategy

  @alias_resolve_strategy.setter
  def alias_resolve_strategy(self, v):
    if v:
      self.__alias_resolver.strategy = v

  @property
  def trace(self):
    return self.__trace

  @property
  def endpoints_resource(self):
    return self.__endpoints_resource

  def with_remote_project(self, key: str):
    if key:
      self.__remote_project_key = ProjectKey(key)

  def with_replay_context(self, context: ReplayContext, replay: Callable[[], Response]): 
    request = context.request
    endpoint = None

    if request.endpoint_id:
      endpoint = self.__get_endpoint(request.endpoint_id)
    elif self.__remote_project_key:
      endpoint = search_endpoint(
        self.endpoints_resource, 
        self.__remote_project_key.id,
        request.method,
        request.url,
        **self.__endpoint_query_params(),
      )

    if endpoint:
      Logger.instance(LOG_ID).debug(f"\tMatched Endpoint: {endpoint}")

      if self.alias_resolve_strategy != alias_resolve_strategy.NONE:
        self.rewrite_request(request, endpoint)

    trace_request = TraceRequest.create(trace_id=self.__trace.id)

    # Set trace id for request
    headers = request.headers
    headers[custom_headers.TRACE_ID] = str(self.trace.id)
    headers[custom_headers.TRACE_REQUEST_ID] = str(trace_request.id)
    request.headers = headers

    # Replay request
    response = replay(context)

    if endpoint:
      self.__create_trace_aliases(trace_request, response, endpoint)

    self.__requests.append((request, response))

  def create_trace_alias(self, alias_name, value, trace_request = None):
    return self.__alias_resolver.create_alias(alias_name, value, trace_request)

  def rewrite_request(self, request: Request, endpoint: EndpointShowResponse):
    if not endpoint:
      return

    id_to_alias = build_id_to_alias_map(endpoint['aliases'])

    self.__rewrite_path(request, endpoint['path_segment_names'], id_to_alias) 
    self.__rewrite_headers(request, endpoint['header_names'], id_to_alias)
    self.__rewrite_query_params(request, endpoint['query_param_names'], id_to_alias)
    self.__rewrite_body_params(request, endpoint['body_param_names'], id_to_alias)

  def __rewrite_path(
    self, request: Request, path_segment_names: List[RequestComponentName], id_to_alias: AliasMap
  ):
    path_segment_strings = request.path_segment_strings
    for path_segment_name in path_segment_names:
      position = path_segment_name.get('position') 

      if position == None:
        continue

      if position > len(path_segment_strings):
        continue

      _alias: Alias = id_to_alias.get(path_segment_name['alias_id'])
      if not _alias:
        continue

      current_value = path_segment_strings[position]
      trace_alias = self.__resolve_and_assign_alias(_alias['name'], current_value)
      if not trace_alias:
        continue

      path_segment_strings[position] = str(trace_alias.value)

    request.path = '/' + '/'.join(path_segment_strings)

  def __rewrite_handler(self, component_type: str, alias_name: str, name: str, value):
    Logger.instance(LOG_ID).info(f"{bcolors.OKCYAN}Rewriting{bcolors.ENDC} {component_type} alias {alias_name} {name} => {value}")

  def __rewrite_headers(
    self, request: Request, header_names: List[RequestComponentName], id_to_alias: AliasMap
  ):
    headers = request.headers
    rewrite_handler = lambda alias_name, name, value: self.__rewrite_handler('header', alias_name, name, value)
    self.__rewrite_components(headers, header_names, id_to_alias, rewrite_handler)
    request.headers = headers

  def __rewrite_query_params(
    self, request: Request, query_param_names: List[RequestComponentName], id_to_alias: AliasMap
  ):
    query_params = request.query_params
    rewrite_handler = lambda alias_name, name, value: self.__rewrite_handler('query param', alias_name, name, value)
    self.__rewrite_components(query_params, query_param_names, id_to_alias, rewrite_handler)
    request.query_params = query_params

  def __rewrite_body_params(
    self, request: Request, body_param_names: List[RequestComponentName], id_to_alias: AliasMap
  ):
    body_params = request.body_params

    if is_traversable(body_params):
      rewrite_params(
        body_params, 
        body_param_names, 
        id_to_alias, 
        self.__alias_resolver,
        handle_after_replace=self.__rewrite_body_params_handler
      )

      request.body_params = body_params

  def __rewrite_body_params_handler(self, name, value, trace_alias: TraceAlias):
    self.__rewrite_handler('body param', trace_alias.name, name, value)
    return self.__alias_resolver.assign_alias(trace_alias, value)

  # Query params and headers
  def __rewrite_components(self, components, component_names, id_to_alias: AliasMap, rewrite_handler):
    delimitter = ',' # Comma is reserved for both query params and headers

    visited = {}
    for component_name in component_names:
      _alias: Alias = id_to_alias.get(component_name['alias_id'])
      if not _alias:
        continue
      
      # Query params and headers can have multiple values, process just once
      name = component_name['name']
      if name in visited:
        continue
      else:
        visited[name] = True

      if name not in components:
        continue

      current_values = components.get_all(name)
      joined_value = delimitter.join(current_values)
      trace_alias = self.__resolve_and_assign_alias(_alias['name'], joined_value)

      if not trace_alias:
        continue

      # Remove all values for the key, name
      components.pop(name)

      # Update with new values
      alias_value = str(trace_alias.value)
      new_values = alias_value.split(delimitter)
      for new_value in new_values:
        components.add(name, new_value)

        if rewrite_handler:
          rewrite_handler(_alias['name'], name, new_value)

  def __create_trace_aliases(self, trace_request: TraceRequest, response: Response, endpoint: EndpointShowResponse):
    '''
    1. Parse all aliased properties from response
    2. Create TraceAlias records for each parsed aliased propert
    '''
    content = decode_response(response.content, response.headers.get('content-type'))
    if not is_traversable(content):
        return Logger.instance(LOG_ID).debug('Skipping creating aliases, content is not traversable')

    id_to_alias = {}
    aliases = endpoint['aliases']
    for _alias in aliases:
      id_to_alias[_alias['id']] = _alias

    response_param_names = endpoint['response_param_names']
    Logger.instance(LOG_ID).debug(f"\tBuilding Trace Aliases from: {response_param_names}")

    for response_param_name in response_param_names:
      try:
        values = self.__query_resolves_response(response_param_name, content)
        Logger.instance(LOG_ID).debug(f"\tValues: {values}")
      except Exception as e:
        Logger.instance(LOG_ID).error(e)
        continue

      _alias: Alias = id_to_alias[response_param_name['alias_id']]
      Logger.instance(LOG_ID).debug(f"\tAlias: {_alias}")

      for value in values:  
        if _alias and value:
          self.__alias_resolver.create_alias(_alias['name'], value, trace_request) 

  def __resolve_and_assign_alias(self, alias_name: str, value) -> Union[TraceAlias, None]:
    trace_alias = self.__alias_resolver.resolve_alias(alias_name, value)

    if trace_alias:
      self.__alias_resolver.assign_alias(trace_alias, value)

      return trace_alias

  def __query_resolves_response(self, response_param_name: ResponseParamName, response: Union[list, dict]) -> list:
    '''
    Return value in response specified by query
    '''

    query = response_param_name['query']
    expression = jmespath.compile(query)
    value = expression.search(response)

    return jmespath.flatten(value, query)

  def __get_endpoint(self, endpoint_id: int) -> Union[EndpointShowResponse, None]:
    res = self.__endpoints_resource.show(
      endpoint_id,
      **self.__endpoint_query_params()
    )

    if res.ok:
      return res.json()

  def __endpoint_query_params(self):
    return {
      'aliases': True,
      'filter': 'alias_id',
      'path_segment_names': True,
      'query_param_names': True,
      'header_names': True,
      'body_param_names': True,
      'response_param_names': True
    }

  def __dump_trace_aliases(self):
    aliases = self.__trace.trace_aliases()

    data = []
    for _alias in aliases:
      data.append({
        'assigned_to': _alias.assigned_to,
        'name': _alias.name,
        'value': _alias.value,
      })

    tabulate_print(data, print_handler=Logger.instance(LOG_ID).debug)



import os
import pdb
import time

from typing import List

from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.logger import bcolors, Logger

from .log_event import LogEvent
from .snapshot_types import DELETE_ACTION, PUT_ACTION, Resource

EVENT_DELIMITTER = "\n"
LOG_ID = 'Log'

class Log():

  def __init__(self, data_dir: DataDir = None):
    data_dir = data_dir or DataDir.instance()
    
    self.__log_file_path = data_dir.snapshots_log_file_path
    self.__history_dir_path = data_dir.snapshots_history_dir_path

    if not os.listdir(self.history_dir_path):
      self.rotate()

  @property
  def events(self):
    events = self.raw_events
    events_count = len(events)

    if events_count == 0:
      return [] 

    return list(map(lambda raw_event: LogEvent(raw_event), events))

  @property
  def history_dir_path(self):
    return self.__history_dir_path

  @property
  def history_files(self):
    files = []

    if not os.path.exists(self.history_dir_path):
      return files

    for file_name in os.listdir(self.history_dir_path):
      full_path = os.path.join(self.history_dir_path, file_name)
      if os.path.isfile(full_path):
        files.append(full_path)

    files.sort()
    return files
    
  @property
  def log_file_path(self):
    return self.__log_file_path

  @property
  def target_events(self):
    events = self.events
    return self.collapse(events)

  @property
  def raw_events(self):
    contents = self.read()
    return self.build_raw_events(contents) 

  @property
  def unprocessed_events(self) -> List[LogEvent]:
    events = self.events

    events_count = len(events)
    if events_count == 0:
      return []

    version = self.version.strip()
    version_uuids = []
    if version:
      version_uuids = version.split(EVENT_DELIMITTER)

    return self.remove_processed_events(events, version_uuids)

  @property
  def version(self):
    version_file_path = DataDir.instance().snapshosts_version_path
    version = ''

    if not os.path.exists(version_file_path):
      return version

    with open(version_file_path, 'r') as fp:
      version = fp.read()

    return version

  @version.setter
  def version(self, v: str):
    version_file_path = DataDir.instance().snapshosts_version_path

    with open(version_file_path, 'w') as fp:
      fp.write(v)

  def append(self, event: str, bucket_interval: int = 300):
    file_path = self.__history_file_path(bucket_interval)

    with open(file_path, 'a') as fp:
      fp.write(event + EVENT_DELIMITTER)

  def delete(self, resource: Resource):
    serialized_event = LogEvent.serialize_delete(resource)
    self.append(serialized_event)

  def generate_version(self, uuids: List[str]):
    return EVENT_DELIMITTER.join(uuids)

  def lock(self, events = ''):
    _events = events or self.read()

    if not _events:
      return

    with open(self.log_file_path, 'w') as fp:
      fp.write(_events + EVENT_DELIMITTER)

  def build_log_events(self, raw_events) -> List[LogEvent]:
    return list(map(lambda raw_event: LogEvent(raw_event), raw_events))

  def next_version(self, last_processed_uuid: str = None):
    uuids = self.uuids()

    if not last_processed_uuid:
      return self.generate_version(uuids)

    for i, uuid in enumerate(uuids):
      if uuid == last_processed_uuid:
        return self.generate_version(uuids[0:i + 1])

    return self.generate_version(uuids)

  def collapse(self, events: List[LogEvent]) -> List[LogEvent]:
    events_count = {}

    # More recent events take precedence over earlier ones, keep only the most recent event 
    for event in events:
      event_key = event.key

      if event_key not in events_count:
        events_count[event_key] = 0

      events_count[event_key] += 1

    target_events = []
    for event in events: 
      event_key = event.key
      events_count[event_key] -= 1

      if events_count[event_key] == 0:
        target_events.append(event)

    return target_events

  def put(self, resource: Resource):
    serialized_event = LogEvent.serialize_put(resource)
    self.append(serialized_event)

  def prune(self, dry_run = False):
    # event uuid => history path
    events: List[LogEvent] = []
    path_index = {}

    history_files = self.history_files
    for file_path in history_files:
      with open(file_path, 'r') as fp:
        contents = fp.read().strip()
        raw_events = self.build_raw_events(contents)
        history_events = self.build_log_events(raw_events)

        for event in history_events:
          path_index[event.uuid] = file_path

        events += history_events
    
    # resource_uuid => event
    resource_index = {}
    for event in events:
      if event.resource_uuid not in resource_index:
        resource_index[event.resource_uuid] = []

      resource_index[event.resource_uuid].append(event)

    pruned_events = self.collapse(events)
    for event in pruned_events:
      snapshot = event.snapshot()
      snapshot_exists = snapshot.exists

      if event.action == DELETE_ACTION or not snapshot_exists:
        Logger.instance(LOG_ID).info(f"{bcolors.OKBLUE}Removing{bcolors.ENDC} {event.resource} {event.resource_uuid}")

        resource_events: List[LogEvent] = resource_index[event.resource_uuid]
        removed_events = {}

        for event in resource_events:
          history_path = path_index[event.uuid]
          if event.uuid in removed_events:
            continue

          Logger.instance(LOG_ID).info(f"Removing event {event.uuid}")
          self.remove_event_history(event, history_path, dry_run)
          removed_events[event.uuid] = True

        if event.action == DELETE_ACTION and snapshot_exists: 
          if not dry_run:
            snapshot.remove()
          Logger.instance(LOG_ID).info(f"Removing {event.resource} snapshot")

  def build_raw_events(self, contents: str) -> List[str]:
    if not contents:
      return []
    events = contents.strip().split(EVENT_DELIMITTER)
    return list(filter(lambda e: not not e, events))
      
  def read(self):
    history_files = self.history_files 

    log = []
    for file_path in history_files:
      with open(file_path, 'r') as fp:
        log.append(fp.read().strip())

    return EVENT_DELIMITTER.join(log)
  
  def remove_dangling_events(self, processed_events: List[LogEvent], unprocessed_events: List[LogEvent]):
    '''
    Remove DELETE events where the last processed event was a PUT
    '''
    index = {}
    for event in processed_events:
      if event.action == PUT_ACTION:
        index[event.resource_uuid] = event
      elif event.action == DELETE_ACTION:
        if event.resource_uuid in index:
          del index[event.resource_uuid]
    
    return list(
      filter(
        lambda e: e.action != DELETE_ACTION or (e.action == DELETE_ACTION and e.resource_uuid in index), 
        unprocessed_events
      )
    )

  def remove_event_history(self, event: LogEvent, history_path: str, dry_run = False):
    events = []
    raw_events = []

    with open(history_path, 'r') as fp:
      contents = fp.read().strip()
      raw_events = self.build_raw_events(contents)
      events = self.build_log_events(raw_events)
      events = list(filter(lambda log_event: log_event.uuid != event.uuid, events))

    if len(events) == 0:
      Logger.instance(LOG_ID).info(f"Removing {history_path}")

      if not dry_run:
        os.remove(history_path)
    else:
      new_raw_events = list(map(lambda event: str(event), events))
      Logger.instance(LOG_ID).info(f"Updating {history_path}, Events: {len(raw_events)} -> {len(new_raw_events)}")

      if not dry_run:
        with open(history_path, 'w') as fp:
          fp.write(EVENT_DELIMITTER.join(new_raw_events))

  def remove_processed_events(self, events, version_uuids):
    i = self.__diverge_point(events, version_uuids)
    processed_events = self.collapse(events[0:i])
    unprocessed_events = self.collapse(events[i:])
    remaining_version_uuids = version_uuids[i:]

    remaining_events = self.remove_dangling_events(processed_events, unprocessed_events)
    return list(filter(lambda e: e.uuid not in remaining_version_uuids, remaining_events))

  # Rotate log to history
  def rotate(self):
    if not os.path.exists(self.__log_file_path):
      return

    with open(self.__log_file_path, 'r') as fp:
      events = fp.read()
      if events:
        self.append(events.strip())

  def uuids(self, events: List[LogEvent] = None):
    _events = events or self.events
    return list(map(lambda e: e.uuid, _events))

  def __diverge_point(self, events: List[LogEvent], version_uuids: List[str]):
    iterations = min(len(events), len(version_uuids))
    for i in range(0, iterations):
      event = events[i] 

      if event.uuid != version_uuids[i]:
        return i
    return iterations

  def __history_file_path(self, bucket_interval: int):
    file_name = f"{int(time.time() / bucket_interval) * bucket_interval}"
    return os.path.join(self.history_dir_path, file_name)


request_show_response = {
  "created_at": "2022-06-08T04:06:58.000Z",
  "endpoint_id": 1308,
  "host": "alpha.api.stoobly.com",
  "http_version": 1.1,
  "id": 11131,
  "latency": 373,
  "method": "GET",
  "path": "/requests/11109/responses/mock",
  "port": "443",
  "position": 7,
  "project_id": 37,
  "query": "project_id=38",
  "scenario_id": 289,
  "scheme": "https",
  "starred": False,
  "status": 200,
  "updated_at": "2022-06-08T04:18:04.995Z",
  "url": "https://alpha.api.stoobly.com/requests/11109/responses/mock?project_id=38",
  "components": [
      "query_params",
      "response"
  ],
  "endpoint": "/requests/:requestId/responses/mock",
  "key": "eyJpIjoxMTEzMSwicCI6MzcsInQiOjU2MTc0MDN9"
}


import copy
import json
import pdb
import pytest
from stoobly_agent.app.cli.helpers.request_synchronize_handler import RequestSynchronizeHandler

from stoobly_agent.app.cli.helpers.request_synchronize_handler import RequestSynchronizeHandler
from stoobly_agent.app.cli.helpers.synchronize_request_service import SynchronizeRequestService
from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.test.mock_data.endpoint_show_response import endpoint_show_response


@pytest.mark.openapi
class TestSynchronizeRequestService():
  @pytest.fixture(scope='class')
  def synchronize_request_service(self) -> SynchronizeRequestService:
    return SynchronizeRequestService(local_db_request_adapter=None)

  @pytest.fixture(scope='class')
  def query_synchronize_handler(self) -> RequestSynchronizeHandler:
    return RequestSynchronizeHandler(request_component.QUERY_PARAM, {})

  @pytest.fixture(scope='class')
  def body_synchronize_handler(self) -> RequestSynchronizeHandler:
    return RequestSynchronizeHandler(request_component.BODY_PARAM, {})

  @pytest.fixture(scope='class')
  def response_synchronize_handler(self) -> RequestSynchronizeHandler:
    return RequestSynchronizeHandler(request_component.RESPONSE_PARAM, {})
  
  @pytest.fixture(scope='function')
  def query_param_names_facade(self) -> RequestComponentNamesFacade:
    query_param_names = copy.deepcopy(endpoint_show_response['query_param_names'])
    aliases = copy.deepcopy(endpoint_show_response['aliases'])

    facade = RequestComponentNamesFacade(query_param_names).with_aliases(aliases)
    return facade
  
  @pytest.fixture(scope='function')
  def body_param_names_facade(self):
    body_param_names = copy.deepcopy(endpoint_show_response['body_param_names'])
    aliases = copy.deepcopy(endpoint_show_response['aliases'])

    facade = RequestComponentNamesFacade(body_param_names).with_aliases(aliases)
    return facade
  
  @pytest.fixture(scope='function')
  def response_param_names_facade(self) -> RequestComponentNamesFacade:
    response_param_names = copy.deepcopy(endpoint_show_response['response_param_names'])
    aliases = copy.deepcopy(endpoint_show_response['aliases'])

    facade = RequestComponentNamesFacade(response_param_names).with_aliases(aliases)
    return facade


  class TestWhenAdding():
    def test_dict_of_object_one_property(self, synchronize_request_service: SynchronizeRequestService, body_param_names_facade: RequestComponentNamesFacade, body_synchronize_handler: RequestSynchronizeHandler):
      self.__decorate_with_values(body_param_names_facade, 'name', ['abc'])
      body_params = {}
      expected_body_params = {'name': 'abc'}

      result = synchronize_request_service.synchronize_component(body_params, body_param_names_facade, body_synchronize_handler)

      assert self.__equals(body_params, expected_body_params), print(body_params)

      success = result[0]
      assert success == True

    def test_dict_does_not_add_extra_optional_params(self, synchronize_request_service: SynchronizeRequestService, query_param_names_facade: RequestComponentNamesFacade, query_synchronize_handler: RequestSynchronizeHandler):
      query_params = {'limit': 3, 'sort': 'color'}
      expected_query_params = {'limit': 3, 'sort': 'color'}

      result = synchronize_request_service.synchronize_component(query_params, query_param_names_facade, query_synchronize_handler)

      assert self.__equals(query_params, expected_query_params), print(query_params)

      success = result[0]
      assert success == True

    def test_array_of_object_one_property(self, synchronize_request_service: SynchronizeRequestService, response_param_names_facade: RequestComponentNamesFacade, response_synchronize_handler: RequestSynchronizeHandler):
      self.__decorate_with_values(response_param_names_facade, '[*].id', [934])
      response = [{}]

      result = synchronize_request_service.synchronize_component(response, response_param_names_facade, response_synchronize_handler)

      expected_response = [{'id': 934}]
      assert self.__equals(response, expected_response), print(response)

      success = result[0]
      assert success == True

    def test_array_of_object_two_property(self, synchronize_request_service: SynchronizeRequestService, response_param_names_facade: RequestComponentNamesFacade, response_synchronize_handler: RequestSynchronizeHandler):
      self.__decorate_with_values(response_param_names_facade, '[*].id', [934])
      self.__decorate_with_values(response_param_names_facade, '[*].category', [1])

      response = [{}]
      result = synchronize_request_service.synchronize_component(response, response_param_names_facade, response_synchronize_handler)

      expected_response = [{'category': 1, 'id': 934}]
      assert self.__equals(response, expected_response), print(response)

      success = result[0]
      assert success == True

    def test_array_of_objects_one_property(self, synchronize_request_service: SynchronizeRequestService, response_param_names_facade: RequestComponentNamesFacade, response_synchronize_handler: RequestSynchronizeHandler):
      self.__decorate_with_values(response_param_names_facade, '[*].id', [934])

      response = [{}, {}]
      result = synchronize_request_service.synchronize_component(response, response_param_names_facade, response_synchronize_handler)

      expected_response = [{'id': 934}, {'id': 934}]
      assert self.__equals(response, expected_response), print(response)

      success = result[0]
      assert success == True

    def __decorate_with_values(self, facade: RequestComponentNamesFacade, query: str, value):
      facade.query_index[query]['values'] = value

    def __equals(self, a, b):
      return json.dumps(a) == json.dumps(b)


  class TestWhenRemoving():
    def test_dict_of_object(self, synchronize_request_service: SynchronizeRequestService, response_param_names_facade: RequestComponentNamesFacade, response_synchronize_handler):
      response = {'to-remove': 1}
      expected_response = {}

      result = synchronize_request_service.synchronize_component(response, response_param_names_facade, response_synchronize_handler)
      success = result[0]

      assert self.__equals(response, expected_response), print(response) 
      assert success == True

    def test_dict_of_objects(self, synchronize_request_service: SynchronizeRequestService, body_param_names_facade: RequestComponentNamesFacade, body_synchronize_handler: RequestSynchronizeHandler):
      self.__decorate_with_values(body_param_names_facade, 'name', ['abc'])
      response = {'name': 'abc', 'to-remove': 1}
      expected_response = {'name': 'abc'}

      result = synchronize_request_service.synchronize_component(response, body_param_names_facade, body_synchronize_handler)
      success = result[0]

      assert self.__equals(response, expected_response), print(response)
      assert success == True
    
    def test_array_of_object_one_property(self, synchronize_request_service: SynchronizeRequestService, response_param_names_facade: RequestComponentNamesFacade, response_synchronize_handler: RequestSynchronizeHandler):
      response = [{'to-remove': 1}]
      result = synchronize_request_service.synchronize_component(response, response_param_names_facade, response_synchronize_handler)

      expected_response = [{}]
      assert self.__equals(response, expected_response), print(response) 

      success = result[0]
      assert success == True

    def test_object_two_properties(self, synchronize_request_service: SynchronizeRequestService, response_param_names_facade: RequestComponentNamesFacade, response_synchronize_handler: RequestSynchronizeHandler):
      response = [{'id': 1, 'remove': 1, 'remove2': 1}]
      synchronize_request_service.synchronize_component(response, response_param_names_facade, response_synchronize_handler)

      assert self.__equals(response, [{'id': 1}]), print(response) 

    def __equals(self, a, b):
      return json.dumps(a) == json.dumps(b)

    def __decorate_with_values(self, facade: RequestComponentNamesFacade, query: str, value):
      facade.query_index[query]['values'] = value




from stoobly_orator.migrations import Migration

class AddTraceRequestReferenceToTraceAliases(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('trace_aliases') as table:
            table.integer('trace_request_id').nullable()
            table.foreign('trace_request_id').references('id').on('trace_requests').index() 

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('trace_aliases') as table:
            table.drop_column('trace_request_id')



import click
import json
import time
import os

from typing import List

from stoobly_agent.app.settings import Settings
from stoobly_agent.app.settings.constants import firewall_action, request_component
from stoobly_agent.app.settings.firewall_rule import FirewallRule
from stoobly_agent.app.settings.match_rule import MatchRule
from stoobly_agent.app.settings.rewrite_rule import ParameterRule, RewriteRule, UrlRule
from stoobly_agent.config.constants import env_vars, mode
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.api.keys import ProjectKey, ScenarioKey
from stoobly_agent.lib.logger import Logger

from .helpers import ProjectFacade, ScenarioFacade
from .helpers.handle_config_update_service import (
    handle_project_update,
    handle_scenario_update,
)
from .helpers.print_service import FORMATS, print_projects, print_scenarios, select_print_options
from .helpers.validations import *

LOG_ID = 'ConfigCLI'

settings = Settings.instance()
is_remote = settings.cli.features.remote or not not os.environ.get(env_vars.FEATURE_REMOTE)

@click.group(
    epilog="Run 'stoobly-agent config COMMAND --help' for more information on a command.",
    help="Manage proxy config"
)
@click.pass_context
def config(ctx):
    pass

@config.command(
    help="Display config contents"
)
@click.option('--dir', is_flag=True, help='To only show the path of the data directory being used.')
@click.option('--save-to-file', is_flag=True, default=False, help='To save to a file or not.')
def dump(**kwargs):
    if kwargs['dir']:
        output = DataDir.instance().path
        print(output)

        return

    settings = Settings.instance()

    output = json.dumps(settings.to_dict(), indent=2, sort_keys=True)

    if kwargs['save_to_file']:
        timestamp = str(int(time.time() * 1000))
        config_dump_file_name = f"stoobly_agent_config_dump_{timestamp}.json"

        with open(config_dump_file_name, 'w') as output_file:
            output_file.write(output)

        print(f"Config successfully dumped to {config_dump_file_name}")
    else:
        print(output)

@config.command(
    help="Reset config to defaults"
)
def reset():
    settings = Settings.instance()
    settings.reset()

    print(f"Reset {settings.path} to defaults.")

### Scenario

@click.group(
    help="Manage active scenario"
)
@click.pass_context
def scenario(ctx):
    pass

@scenario.command(
    help="Describe scenario"
)
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
def show(**kwargs):
    settings = Settings.instance()
    print_options = select_print_options(kwargs)

    project_key = __project_key(settings)
    data_rule = settings.proxy.data.data_rules(project_key.id)

    if not data_rule.scenario_key or len(data_rule.scenario_key) == 0:
        return

    kwargs['key'] = data_rule.scenario_key

    scenario_key = resolve_scenario_key_and_validate(kwargs, settings)
    scenario = ScenarioFacade(settings)

    try:
        scenario_response = scenario.show(scenario_key)
    except AssertionError as e:
        return print(e, file=sys.stderr)

    print_scenarios([scenario_response], **print_options)

@scenario.command(
    help="Set active scenario."
)
@click.argument('scenario_key')
def set(**kwargs):
    validate_scenario_key(kwargs['scenario_key'])
    scenario_key = ScenarioKey(kwargs['scenario_key'])

    settings = Settings.instance()
    project_key = __project_key(settings)

    if scenario_key.project_id != project_key.id:
        return print("Please provide a scenario that belongs to the current project.\n")

    data_rule = settings.proxy.data.data_rules(project_key.id)

    data_rule.scenario_key = kwargs['scenario_key']

    handle_scenario_update(settings)

    settings.commit()

    print("Scenario updated!")

@scenario.command(
    help="Clear active scenario."
)
def clear(**kwargs):
    settings = Settings.instance()

    project_key = __project_key(settings) 

    data_rule = settings.proxy.data.data_rules(project_key.id)
    data_rule.scenario_key = ''
    settings.commit()

    print("Scenario cleared!")

### Rewrite

@click.group(
    help="Manage rewrite rules"
)
@click.pass_context
def rewrite(ctx):
    pass

@rewrite.command(
    help="Set rewrite rule."
)
@click.option('--hostname', help='Request URL hostname.')
@click.option(
    '--method', 
    multiple=True, 
    required=True,
    type=click.Choice(['GET', 'POST', 'DELETE', 'OPTIONS', 'PUT']), 
    help='HTTP methods.'
)
@click.option(
    '--mode',
    multiple=True,
    required=True,
    type=click.Choice([mode.MOCK, mode.RECORD, mode.REPLAY] + ([mode.TEST] if is_remote else []))
)
@click.option('--name', help='Name of the request component.')
@click.option('--path', help='Request URL path to rewrite to.')
@click.option('--pattern', required=True, help='URLs to rewrite.')
@click.option('--port', help='Request URL port to rewrite to.')
@click.option('--project-key', help='Project to add rewrite rule to.')
@click.option(
    '--type', 
    type=click.Choice([request_component.BODY_PARAM, request_component.HEADER, request_component.QUERY_PARAM, request_component.RESPONSE_HEADER, request_component.RESPONSE_PARAM]), 
    help='Request component type.'
)
@click.option(
    '--scheme', 
    type=click.Choice(['http', 'https']), 
    help='Request URL scheme to rewrite to.'
)
@click.option('--value', help='Rewrite value.')
def set(**kwargs):
    if kwargs['name'] or kwargs['value'] or kwargs['type']:
        if kwargs['name'] == None:
            print("Error: Missing option '--name'", file=sys.stderr)
            sys.exit(1)

        if kwargs['value'] == None:
            print("Error: Missing option '--value'", file=sys.stderr)
            sys.exit(1)
        
        if kwargs['type'] == None:
            print("Error: Missing option '--type'", file=sys.stderr)
            sys.exit(1)

    settings = Settings.instance()
    project_key_str = resolve_project_key_and_validate(kwargs, settings)
    project_key = ProjectKey(project_key_str)

    methods = list(kwargs['method'])
    modes = list(kwargs['mode'])

    rewrite_rules = settings.proxy.rewrite.rewrite_rules(project_key.id)

    rewrite_rule_filter = lambda rule: rule.pattern == kwargs['pattern'] and rule.methods == methods
    filtered_rewrite_rules: List[RewriteRule] = list(filter(rewrite_rule_filter, rewrite_rules))

    if len(filtered_rewrite_rules) == 0:
        parameter_rules = list(filter(lambda r: r != None, [__select_parameter_rule(kwargs)]))
        url_rules = list(filter(lambda r: r != None , [__select_url_rule(kwargs)]))
        rewrite_rule = RewriteRule({
            'methods': methods,
            'pattern': kwargs['pattern'],
            'parameter_rules': parameter_rules,
            'url_rules': url_rules,
        })
        rewrite_rules.append(rewrite_rule)
        settings.proxy.rewrite.set_rewrite_rules(project_key.id, rewrite_rules)
    else:
        parameter_rule_filter = lambda rule: rule.name == kwargs['name'] and rule.type == kwargs['type'] and rule.modes == modes
        url_rule_filter = lambda rule: rule.modes == modes

        for rewrite_rule in filtered_rewrite_rules:
            # Parameter rules
            parameter_rule_dict = __select_parameter_rule(kwargs)

            if parameter_rule_dict:
                parameter_rules = rewrite_rule.parameter_rules
                filtered_parameter_rules: List[ParameterRule] = list(filter(parameter_rule_filter, parameter_rules))

                if len(filtered_parameter_rules) == 0:
                    parameter_rule = ParameterRule(parameter_rule_dict)
                    parameter_rules.append(parameter_rule)
                    rewrite_rule.parameter_rules = parameter_rules
                else:
                    for parameter_rule in filtered_parameter_rules:
                        parameter_rule.update(parameter_rule_dict)

            # URL rules
            url_rule_dict = __select_url_rule(kwargs)
            
            if url_rule_dict:
                url_rules = rewrite_rule.url_rules
                filtered_url_rules: List[UrlRule] = list(filter(url_rule_filter, url_rules))

                if len(filtered_url_rules) == 0:
                    url_rule = UrlRule(url_rule_dict)
                    url_rules.append(url_rule)
                    rewrite_rule.url_rules = url_rules
                else:
                    for url_rule in filtered_url_rules:
                        url_rule.update(url_rule_dict)

    settings.commit()

    Logger.instance(LOG_ID).debug(f"Rewrite {kwargs['name']} -> {kwargs['value']} set!")

### Match

@click.group(
    help="Manage match rules"
)
@click.pass_context
def match(ctx):
    pass

@match.command(
    help="Set match rule."
)
@click.option(
    '--method', 
    multiple=True, 
    required=True,
    type=click.Choice(['GET', 'POST', 'DELETE', 'OPTIONS', 'PUT']), 
    help='HTTP methods.'
)
@click.option(
    '--mode',
    multiple=True,
    required=True,
    type=click.Choice([mode.MOCK] + ([mode.TEST] if is_remote else []))
)
@click.option('--pattern', required=True, help='URLs pattern.')
@click.option('--project_key', help='Project to add rewrite rule to.')
@click.option(
    '--component', 
    multiple=True,
    type=click.Choice([request_component.BODY_PARAM, request_component.HEADER, request_component.QUERY_PARAM]), 
    help='Request component type.'
)
def set(**kwargs):
    settings = Settings.instance()
    project_key_str = resolve_project_key_and_validate(kwargs, settings)
    project_key = ProjectKey(project_key_str)

    methods = list(kwargs['method'])
    modes = list(kwargs['mode'])
    match_rule = MatchRule({
        'components': list(kwargs['component']),
        'methods': methods,
        'modes': modes,
        'pattern': kwargs['pattern'],
    })

    match_rules = settings.proxy.match.match_rules(project_key.id)
    index = -1
    i = 0
    for rule in match_rules:
        if rule.pattern == kwargs['pattern'] and rule.methods == methods:
            index = i
        i += 1

    if index == -1:
        match_rules.append(match_rule)
        settings.proxy.match.set_match_rules(project_key.id, match_rules)
    else:
        match_rules[index] = match_rule

    settings.commit()

    Logger.instance(LOG_ID).debug(f"Match {kwargs['method']} {kwargs['pattern']} -> {kwargs['component']} set!")

### Firewall

@click.group(
    help="Manage firewall rules"
)
@click.pass_context
def firewall(ctx):
    pass

@firewall.command(
    help="Set firewall rule."
)
@click.option(
    '--method', 
    multiple=True, 
    required=True,
    type=click.Choice(['GET', 'POST', 'DELETE', 'OPTIONS', 'PUT']), 
    help='HTTP methods.'
)
@click.option(
    '--mode',
    multiple=True,
    required=True,
    type=click.Choice([mode.MOCK, mode.RECORD, mode.REPLAY] + ([mode.TEST] if is_remote else []))
)
@click.option('--pattern', required=True, help='URLs pattern.')
@click.option('--project_key', help='Project to add firewall rule to.')
@click.option(
    '--action', 
    required=True,
    type=click.Choice([firewall_action.EXCLUDE, firewall_action.INCLUDE]), 
    help='Action to take on match.'
)
def set(**kwargs):
    settings = Settings.instance()
    project_key_str = resolve_project_key_and_validate(kwargs, settings)
    project_key = ProjectKey(project_key_str)

    methods = list(kwargs['method'])
    modes = list(kwargs['mode'])
 
    firewall_rule = FirewallRule({
        'action': kwargs['action'],
        'methods': methods,
        'modes': modes,
        'pattern': kwargs['pattern'],
    })

    firewall_rules = settings.proxy.firewall.firewall_rules(project_key.id)
    index = -1
    i = 0
    for rule in firewall_rules:
        if rule.pattern == kwargs['pattern'] and rule.methods == methods:
            index = i
        i += 1

    if index == -1:
        firewall_rules.append(firewall_rule)
        settings.proxy.firewall.set_firewall_rules(project_key.id, firewall_rules)
    else:
        firewall_rules[index] = firewall_rule

    settings.commit()

    Logger.instance(LOG_ID).debug(f"Firewall {kwargs['method']} {kwargs['pattern']} -> {kwargs['action']} set!")

### Validate

@config.command(
    help="Check config validity"
)
def validate(**kwargs):
    Settings.instance().validate()

if is_remote:
    ### API Key

    @click.group(
        help="Manage API key"
    )
    @click.pass_context
    def api_key(ctx):
        pass

    @api_key.command(
        help="Set API Key"
    )
    @click.argument('api_key')
    def set(**kwargs):
        settings = Settings.instance()

        api_key = kwargs['api_key']
        settings.remote.api_key = api_key

        settings.commit()

        print("API Key updated!")

    @click.group(
        help="Manage active project"
    )
    @click.pass_context
    def project(ctx):
        pass

    @project.command(
        help="Use local project."
    )
    def local(**kwargs):
        project_key = ProjectKey.local_key
        __set_project_key(project_key)

        print("Using local project!")

    @project.command(
        help="Describe project."
    )
    @click.option('--format', type=click.Choice(FORMATS), help='Format output.')
    @click.option('--select', multiple=True, help='Select column(s) to display.')
    @click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
    def show(**kwargs):
        settings = Settings.instance()
        print_options = select_print_options(kwargs)

        project_key = __project_key(settings)
        if project_key.is_local:
            return print('Using local project')

        kwargs['project_key'] = project_key.raw

        project_key = resolve_project_key_and_validate(kwargs, settings)
        project = ProjectFacade(settings)

        try:
            project_response = project.show(project_key)
        except AssertionError as e:
            return print(e, file=sys.stderr)

        print_projects([project_response], **print_options)

    @project.command(
        help="Set active project."
    )
    @click.argument('project_key')
    def set(**kwargs):
        project_key = kwargs['project_key']

        __set_project_key(project_key)

        print("Project updated!")

if is_remote:
    config.add_command(api_key)

config.add_command(firewall)
config.add_command(match)

if is_remote:
    config.add_command(project)

config.add_command(rewrite)
config.add_command(scenario)

def __select_parameter_rule(kwargs):
    if kwargs['name'] == None or kwargs['value'] == None or kwargs['type'] == None:
        return

    return {
        'modes': list(kwargs['mode']),
        'name': kwargs['name'],
        'value': kwargs['value'],
        'type': kwargs['type'],
    }

def __select_url_rule(kwargs):
    if  kwargs['scheme'] == None and kwargs['hostname'] == None and kwargs['port'] == None and kwargs['path'] == None:
        return

    return {
        'hostname': kwargs['hostname'],
        'modes': list(kwargs['mode']),
        'path': kwargs['path'],
        'port': kwargs['port'],
        'scheme': kwargs['scheme'],
    }

def __project_key(settings):
    project_key = settings.proxy.intercept.project_key
    validate_project_key(project_key)
    return ProjectKey(project_key)

def __set_project_key(project_key: str):
    validate_project_key(project_key)

    settings = Settings.instance()
    settings.proxy.intercept.project_key = project_key 

    handle_project_update(settings)

    settings.commit()



from stoobly_orator.migrations import Migration

class CreateTraceRequests(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create('trace_requests') as table:
            table.increments('id')
            table.integer('trace_id').unsigned()
            table.foreign('trace_id').references('id').on('traces').index()
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop('trace_requests')



import json

from stoobly_orator.orm import accessor, belongs_to

from .base import Base

DICT_TYPE = type({}).__name__
LIST_TYPE = type([]).__name__

class TraceAlias(Base):
  __fillable__ = ['assigned_to', 'assigned_to_inferred_type', 'name', 'trace_id', 'trace_request_id', 'value', 'value_inferred_type']

  @belongs_to
  def trace(self):
    from .trace import Trace # Get around circular import
    return Trace

  @belongs_to
  def trace_request(self):
    from .trace_request import TraceRequest
    return TraceRequest

  @accessor
  def assigned_to(self):
    _assigned_to = self.raw_assigned_to

    return self.__deserialize(_assigned_to, self.assigned_to_inferred_type)

  @accessor
  def value(self):
    _value = self.raw_value

    return self.__deserialize(_value, self.value_inferred_type)

  @property
  def raw_assigned_to(self):
    try:
      return self.get_raw_attribute('assigned_to')
    except KeyError as e:
      return None

  @property
  def raw_value(self):
    try:
      return self.get_raw_attribute('value')
    except KeyError as e:
      return None

  # Override
  def to_dict(self):
    return super().attributes_to_dict()

  def serialize_assigned_to(self):
    _assigned_to = self.raw_assigned_to
    assigned_to_inferred_type = type(_assigned_to).__name__

    # If non primitive type, serialize
    if assigned_to_inferred_type == DICT_TYPE or assigned_to_inferred_type == LIST_TYPE:
      self.assigned_to = json.dumps(_assigned_to)

    self.assigned_to_inferred_type = assigned_to_inferred_type

  def serialize_value(self):
    _value = self.raw_value
    value_inferred_type = type(_value).__name__

    # If non primitive type, serialize
    if value_inferred_type == DICT_TYPE or value_inferred_type == LIST_TYPE:
      self.value = json.dumps(_value)

    self.value_inferred_type = value_inferred_type

  def __deserialize(self, value, inferred_type):
    if inferred_type == DICT_TYPE or inferred_type == LIST_TYPE:
      # If updating, _value can be non str or byte. _value will be what the user sets
      if isinstance(value, bytes) or isinstance(value, str):
        return json.loads(value)
      else:
        return value
    else:
      return value

# Handles creating and updating
def handle_saving(trace_alias):
  trace_alias.serialize_value()
  trace_alias.serialize_assigned_to()

TraceAlias.saving(handle_saving)



import pdb
import pytest

from stoobly_agent.mock import Mock
from stoobly_agent.app.models.schemas.request import Request

@pytest.fixture
def request_with_body_headers_query_params_response():
  res = Mock.get('http://0.0.0.0:3002/requests/14812?project_id=67&body=True&headers=True&query_params=True&response=True')
  return res.json()

@pytest.fixture
def stoobly_request(request_with_body_headers_query_params_response):
  return Request(request_with_body_headers_query_params_response)

class TestRequest():
  
  def test_url(self, stoobly_request: Request):
    assert stoobly_request.url == 'http://www.google.com/'

  def test_path(self, stoobly_request: Request):
    assert stoobly_request.path == '/'

  def test_method(self, stoobly_request: Request):
    assert stoobly_request.method == 'GET'

  def test_query_params(self, stoobly_request: Request):
    assert len(stoobly_request.query_params) == 0

  def test_body(self, stoobly_request: Request):
    assert stoobly_request.body == ''

  def test_headers(self, stoobly_request: Request):
    assert stoobly_request.headers['content-length'] == '0'


import importlib
import os
import pdb
import pytest
import tempfile

from click.testing import CliRunner

from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.cli import record, request, snapshot
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.api.keys import RequestKey
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestCopy():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
      return reset()

    class TestWhenRequest():
      @pytest.fixture(scope='class')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(autouse=True, scope='class')
      def snapshot_result(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0
        return snapshot_result

      def test_it_copies(self, runner: CliRunner, recorded_request: Request):
        with tempfile.TemporaryDirectory() as dirpath:
          data_dir = DataDir(dirpath)
          log = Log(data_dir)

          # Check that no events have been copied
          assert len(log.events) == 0

          apply_result = runner.invoke(snapshot, ['copy', dirpath, '--request-key', recorded_request.key()])
          assert apply_result.exit_code == 0

          # Check that one event has been copied
          assert len(log.events) == 1

          event = log.events[0]
          request_key = RequestKey(recorded_request.key())

          # Check that the event is the request
          assert request_key.id == event.resource_uuid


import os

from .feature_settings import FeatureSettings
from .types.cli_settings import CLISettings as ICLISettings 

class CLISettings():

  def __init__(self, cli_settings: ICLISettings):
    self.__cli_settings = cli_settings

    self.__features = FeatureSettings(self.__cli_settings.get('features'))

  @property
  def features(self) -> FeatureSettings:
    return self.__features

  def to_dict(self) -> ICLISettings:
    return {
      'features': self.__features.to_dict()
    }


from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings

class InterceptContext():

  def __init__(self, flow: MitmproxyHTTPFlow, intercept_settings: InterceptSettings):
    self.__flow = flow
    self.__intercept_settings = intercept_settings

  @property
  def flow(self):
    return self.__flow

  @flow.setter
  def flow(self, v: MitmproxyHTTPFlow):
    self.__flow = v

  @property
  def intercept_settings(self):
    return self.__intercept_settings




from stoobly_orator.migrations import Migration

class AddScenarioIdReferenceToRequests(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('requests') as table:
            table.integer('scenario_id').nullable()
            table.foreign('scenario_id').references('id').on('scenarios').index() 

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('requests') as table:
            table.drop_column('scenario_id')



import pdb
import pytest

from click.testing import CliRunner
from typing import List

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.settings import Settings
from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.app.settings.rewrite_rule import ParameterRule, RewriteRule, UrlRule
from stoobly_agent.cli import config
from stoobly_agent.config.constants import mode
from stoobly_agent.lib.api.keys.project_key import LOCAL_PROJECT_ID

@pytest.fixture(scope='class')
def settings():
  return reset()

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

class TestRewrite():

  class TestParameter():

    @pytest.fixture(scope='class')
    def header_name(self):
      return 'access-token'

    @pytest.fixture(scope='class')
    def header_value(self):
      return ''

    @pytest.fixture(scope='class')
    def method(self):
      return 'GET'

    @pytest.fixture(scope='class')
    def _mode(self):
      return mode.RECORD

    @pytest.fixture(scope='class')
    def pattern(self):
      return '.*?'

    @pytest.fixture(scope='class')
    def _type(self):
      return request_component.HEADER

    @pytest.fixture(autouse=True, scope='class')
    def rewrite_result(self, settings: Settings, runner: CliRunner, method, _mode, header_name, header_value, pattern, _type):
      rewrite_result = runner.invoke(config, [
          'rewrite', 'set', 
          '--method', method, '--mode', _mode, '--name', header_name, '--value', header_value, '--pattern', pattern, '--type', _type
        ]
      )
      settings.load()
      return rewrite_result

    @pytest.fixture()
    def rewrite_rules(self, settings: Settings):
      return settings.proxy.rewrite.rewrite_rules(str(LOCAL_PROJECT_ID))

    @pytest.fixture()
    def rewrite_rule(self, rewrite_rules: List[RewriteRule]):
      return rewrite_rules[0]

    @pytest.fixture()
    def parameter_rule(self, rewrite_rule: RewriteRule):
      parameter_rules = rewrite_rule.parameter_rules
      return parameter_rules[0]

    def test_no_errors_returned(self, rewrite_result):
      assert rewrite_result.exit_code == 0

    def test_one_rewrite_rule(self, rewrite_rules: List[RewriteRule]):
      assert len(rewrite_rules) == 1

    def test_pattern(self, rewrite_rule: RewriteRule, pattern):
      assert rewrite_rule.pattern == pattern

    def test_methods(self, rewrite_rule: RewriteRule, method):
      assert rewrite_rule.methods == [method]

    def test_one_parameter_rule(self, rewrite_rule: RewriteRule):
      parameter_rules = rewrite_rule.parameter_rules
      assert len(parameter_rules) == 1

    def test_modes(self, parameter_rule: ParameterRule, _mode):
      assert parameter_rule.modes == [_mode]

    def test_type(self, parameter_rule: ParameterRule, _type):
      assert parameter_rule.type == _type

    def test_header_name(self, parameter_rule: ParameterRule, header_name):
      assert parameter_rule.name == header_name

    def test_header_value(self, parameter_rule: ParameterRule, header_value):
      assert parameter_rule.value == header_value

  class TestUrl():

    @pytest.fixture(scope='class')
    def scheme(self):
      return 'https'

    @pytest.fixture(scope='class')
    def hostname(self):
      return 'localhostname'

    @pytest.fixture(scope='class')
    def port(self):
      return '80'

    @pytest.fixture(scope='class')
    def method(self):
      return 'GET'

    @pytest.fixture(scope='class')
    def _mode(self):
      return mode.RECORD

    @pytest.fixture(scope='class')
    def path(self):
      return '/index.html'

    @pytest.fixture(scope='class')
    def pattern(self):
      return '.*?'

    @pytest.fixture(autouse=True, scope='class')
    def rewrite_result(self, settings: Settings, runner: CliRunner, method, _mode, path, pattern, scheme, hostname, port):
      rewrite_result = runner.invoke(config, [
          'rewrite', 'set', 
          '--method', method, '--mode', _mode, '--pattern', pattern, '--hostname', hostname, '--path', path, '--port', port, '--scheme', scheme,
        ]
      )
      settings.load()
      return rewrite_result

    @pytest.fixture()
    def rewrite_rules(self, settings: Settings):
      return settings.proxy.rewrite.rewrite_rules(str(LOCAL_PROJECT_ID))

    @pytest.fixture()
    def rewrite_rule(self, rewrite_rules: List[RewriteRule]):
      return rewrite_rules[0]

    @pytest.fixture()
    def url_rule(self, rewrite_rule: RewriteRule):
      url_rules = rewrite_rule.url_rules
      return url_rules[0]

    def test_no_errors_returned(self, rewrite_result):
      assert rewrite_result.exit_code == 0

    def test_one_rewrite_rule(self, rewrite_rules: List[RewriteRule]):
      assert len(rewrite_rules) == 1

    def test_pattern(self, rewrite_rule: RewriteRule, pattern):
      assert rewrite_rule.pattern == pattern

    def test_methods(self, rewrite_rule: RewriteRule, method):
      assert rewrite_rule.methods == [method]

    def test_one_url_rule(self, rewrite_rule: RewriteRule):
      url_rules = rewrite_rule.url_rules
      assert len(url_rules) == 1

    def test_modes(self, url_rule: UrlRule, _mode):
      assert url_rule.modes == [_mode]

    def test_url_hostname(self, url_rule: UrlRule, hostname):
      assert url_rule.hostname == hostname

    def test_url_path(self, url_rule: UrlRule, path: str):
      assert url_rule.path == path

    def test_url_port(self, url_rule: UrlRule, port):
      assert url_rule.port == port

    def test_url_scheme(self, url_rule: UrlRule, scheme):
      assert url_rule.scheme == scheme


import json
import pdb

from stoobly_agent.lib.utils import jmespath

class TestReplace():
    def test_a(self):
        a = {'foo': 1}
        jmespath.search('foo', a, { 'replacements': ['*'] })
        assert self.equals(a, {'foo': '*'}), print(a)

    def test_b(self):
        a = {'foo': { 'bar': 1 }}
        jmespath.search('foo.bar', a, { 'replacements': ['*'] })
        assert self.equals(a, {'foo': { 'bar': '*' }}), print(a)

    def test_c(self):
        a = [1]
        jmespath.search('[*]', a, { 'replacements': ['*'] })
        assert self.equals(a, ['*']), print(a)

    def test_d(self):
        a = [[1]]
        jmespath.search('[0][0]', a, { 'replacements': ['*'] })
        assert self.equals(a, ([['*']])), print(a)

    def test_e(self):
        a = [{'a': 1}]
        jmespath.search('[*].a', a, { 'replacements': ['*'] })
        assert self.equals(a, [{'a': '*'}]), print(a)

    def test_f(self):
        a = [{'a': [{'a': 1}]}]
        jmespath.search('[*].a[*].a', a, { 'replacements': ['*'] })
        assert self.equals(a, [{'a': [{'a': '*'}]}]), print(a)

    def test_g(self):
        a = [{'a': [{'a': {'a': [{'a': 1}]}}]}]
        jmespath.search('[*].a[*].a.a[*].a', a, { 'replacements': ['*'] })
        assert self.equals(a, [{'a': [{'a': {'a': [{'a': '*'}]}}]}]), print(a)

    def test_h(self):
        a = [{'a': [{'a': {'a': [{'a': 1}, {'a': 2}]}}]}]
        jmespath.search('[*].a[*].a.a[*].a', a, { 'replacements': ['*'] })
        assert self.equals(a, [{'a': [{'a': {'a': [{'a': '*'}, {'a': '*'}]}}]}]), print(a)

    def test_i(self):
        a = [{'a': 1}, {'a': 2}, {'a': 3}]
        jmespath.search('[*].a', a, { 'replacements': ['*'] })
        assert self.equals(a, [{'a': '*'}, {'a': '*'}, {'a': '*'}]), print(a)

    def test_j(self):
        a = [{'a': 1}, {'a': 'a'}, {'a': None}]
        jmespath.search('[*].a', a, { 'replacements': ['*'] })
        assert self.equals(a, [{'a': '*'}, {'a': '*'}, {'a': '*'}]), print(a)

    def test_k(self):
        a = {'a': [{'a': {'a': 1}}, {'a': {'a': 2}}]}
        jmespath.search('a[*].a.a', a, { 'replacements': ['*'] })
        assert self.equals(a, {'a': [{'a': {'a': '*'}}, {'a': {'a': '*'}}]}), print(a)

    def test_l(self):
        a = {'a': [{'a': {'a': 1}}, {'a': {'a': 2}}, {'a': {'a': 3}}, {'a': {'a': 4}}]}
        jmespath.search('a[*].a.a', a, { 'replacements': ['*', '**', '***', '****'] })
        assert self.equals(a, {'a': [{'a': {'a': '*'}}, {'a': {'a': '**'}}, {'a': {'a': '***'}}, {'a': {'a': '****'}}], }), print(a)

    def test_m(self):
        a = {'a': [[{'a': 1}], [{'a': 2}]]}
        jmespath.search('a[*][*].a', a, { 'replacements': ['*'] })
        assert self.equals(a, {'a': [[{'a': '*'}], [{'a': '*'}]]}), print(a)

    def test_n(self):
        a = {'access-token': '123'}
        jmespath.search('"access-token"', a, { 'replacements': ['']})
        assert self.equals(a, {'access-token': ''}), print(a)

    def test_o(self):
        a = {'a': 1}
        jmespath.search('a.b', a, { 'replacements': ['']})
        assert self.equals(a, {'a': 1}), print(a)

    def equals(self, a, b):
        return json.dumps(a) == json.dumps(b)

class TestSearch():
    def test_a(self):
        a = [{'foo': [1, [2], [3]]}, {'foo': [1, [2], [3]]}]
        res = jmespath.search('[*].foo[*][*]', a)
        assert self.equals(res, [[[2], [3]], [[2], [3]]]), print(a)

    def equals(self, a, b):
        return json.dumps(a) == json.dumps(b)


import pdb

import requests

from stoobly_agent.app.models.schemas.request import Request
from stoobly_agent.app.settings import Settings

from .factories.resource.request import RequestResourceFactory
from .model import Model
from .types import RequestCreateParams, RequestDestroyParams, RequestDestroyAllParams, RequestIndexSimilarParams, RequestShowParams

class RequestModel(Model):

  def __init__(self, settings: Settings, **options):
    super().__init__(settings, **options)

  def as_local(self):
    self.adapter = RequestResourceFactory(self.settings.remote).local_db()
    self.is_local = True

  def as_remote(self):
    self.adapter = RequestResourceFactory(self.settings.remote).stoobly()
    self.is_local = False

  def create(self, **body_params: RequestCreateParams):
    try:
      return self.adapter.create(**body_params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def show(self, request_id: str, **params: RequestShowParams):
    try:
      return self.adapter.show(request_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def index_similar(self, params: RequestIndexSimilarParams):
    try:
      local_adapter = RequestResourceFactory(self.settings.remote).local_db()
      return local_adapter.similar(params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def response(self, **query_params):
    return self.adapter.response(**query_params)

  def index(self, **query_params):
    try:
      return self.adapter.index(**query_params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def update(self, request_id: str, **params: Request):
    try:
      return self.adapter.update(request_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def destroy(self, request_id: str, **params: RequestDestroyParams):
    try:
      return self.adapter.destroy(request_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def destroy_all(self, **params: RequestDestroyAllParams):
    try:
      return self.adapter.destroy_all(**params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def snapshot(self, request_id: str, **params):
    try:
      return self.adapter.snapshot(request_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)



import os
import pdb
import requests

from time import time
from typing import Callable, TypedDict, Union

from stoobly_agent.app.cli.helpers.certificate_authority import CertificateAuthority
from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.models.adapters.python import PythonResponseAdapterFactory
from stoobly_agent.app.models.schemas.request import Request
from stoobly_agent.app.models.replayed_response_model import ReplayedResponseModel
from stoobly_agent.app.proxy.replay.trace_context import TraceContext
from stoobly_agent.app.proxy.simulate_intercept_service import simulate_intercept
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import alias_resolve_strategy, custom_headers, request_origin, test_filter, test_strategy
from stoobly_agent.config.mitmproxy import MitmproxyConfig
from stoobly_agent.config.constants import mock_policy, mode

class ReplayRequestOptions(TypedDict):
  alias_resolve_strategy: alias_resolve_strategy.AliasResolveStrategy
  group_by: str
  host: str
  lifecycle_hooks_path: str
  mode: Union[mode.MOCK, mode.RECORD, mode.TEST, None]
  before_replay: Union[Callable[[ReplayContext], None], None]
  after_replay: Union[Callable[[ReplayContext], Union[requests.Response, None]], None]
  project_key: Union[str, None]
  proxies: dict
  public_directory_path: str
  remote_project_key: str
  report_key: Union[str, None] 
  request_origin: Union[request_origin.CLI, None] 
  response_fixtures_path: str
  response_mode: Union[mode.RECORD, None]
  scenario_key: Union[str, None] 
  scheme: str
  test_filter: test_filter.TestFilter
  test_save_results: bool
  test_strategy: Union[test_strategy.CUSTOM, test_strategy.DIFF, test_strategy.FUZZY]
  trace_context: TraceContext

def replay_with_trace(context: ReplayContext, trace_context: TraceContext, options: ReplayRequestOptions):
  trace_context.alias_resolve_strategy = options.get('alias_resolve_strategy')
  trace_context.with_remote_project(options.get('remote_project_key'))
  return trace_context.with_replay_context(context, lambda context: replay(context, options))

def replay(context: ReplayContext, options: ReplayRequestOptions) -> requests.Response:
  request = context.request
  headers = request.headers
  method = request.method

  if options.get('host'):
    request.host = options['host']
    headers['host'] = options['host']

  if options.get('scheme'):
    request.scheme = options['scheme']

  if options.get('alias_resolve_strategy'):
    headers[custom_headers.ALIAS_RESOLVE_STRATEGY] = options['alias_resolve_strategy']

  if options.get('lifecycle_hooks_path'):
    __handle_path_header(custom_headers.LIFECYCLE_HOOKS_PATH, options['lifecycle_hooks_path'], headers) 

  if options.get('mode'):
    __handle_mode_option(options['mode'], request, headers)

  if options.get('project_key'):
    headers[custom_headers.PROJECT_KEY] = options['project_key']

  if options.get('public_directory_path'):
    __handle_path_header(custom_headers.PUBLIC_DIRECTORY_PATH, options['public_directory_path'], headers)

  if options.get('report_key'):
    headers[custom_headers.REPORT_KEY] = options['report_key']

  if options.get('request_origin'):
    headers[custom_headers.REQUEST_ORIGIN] = options['request_origin']

  if custom_headers.REQUEST_ORIGIN not in headers:
    headers[custom_headers.REQUEST_ORIGIN] = request_origin.CLI

  if options.get('response_fixtures_path'):
    __handle_path_header(custom_headers.RESPONSE_FIXTURES_PATH, options['response_fixtures_path'], headers)

  if options.get('response_mode'):
    headers[custom_headers.RESPONSE_PROXY_MODE] = options['response_mode']

  if options.get('scenario_key'):
    headers[custom_headers.SCENARIO_KEY] = options['scenario_key']

  if options.get('test_filter'):
    headers[custom_headers.TEST_FILTER] = options['test_filter']

  if options.get('test_strategy'):
    headers[custom_headers.TEST_STRATEGY] = options['test_strategy']

  if options.get('remote_project_key'):
    headers[custom_headers.REMOTE_PROJECT_KEY] = options['remote_project_key']

  if options.get('test_save_results'):
    headers[custom_headers.TEST_SAVE_RESULTS] = '1'

  if 'before_replay' in options and callable(options['before_replay']):
    options['before_replay'](context)

  request_config = {
    'allow_redirects': True,
    'stream': True,
    'verify': not MitmproxyConfig.instance().get('ssl_insecure')
  }
  request_dict = {
    'data': request.body,
    'headers': headers, 
    #'params': request.query_params, # Do not send query params, they should be a part of the URL
  }

  now = time()
  res: requests.Response = None
  if not options.get('proxy'):
    request = requests.Request(**{
      **request_dict,
      'method': method,
      'url': request.url,
    })

    res = simulate_intercept(request, **request_config)
  else:
    settings = Settings.instance()
    handler = getattr(requests, method.lower())
    ca = CertificateAuthority()

    res = handler(
      request.url, 
      **{
        **request_config,
        **request_dict,
        'proxies': {
          'http': settings.proxy.url,
          'https': settings.proxy.url,
        },
        'verify': ca.ca_cert_path('.crt') if request_config['verify'] else False,
      }
    )
  received_at = time()
  latency = int((received_at - now) * 1000)

  if options['mode'] == mode.REPLAY:
    if options.get('save') or options.get('overwrite'):
      replayed_response, status = __create_replayed_response(context.request.id, res, latency)

      if status < 400 and options.get('overwrite'):
        __overwrite_response(replayed_response.get('id'))

  if 'after_replay' in options and callable(options['after_replay']):
    context.with_response(res)
    res = options['after_replay'](context) or res

  return res

def __handle_path_header(header_name: str, script_path: str, headers):
  if not script_path:
    return

  if not os.path.isabs(script_path):
    script_path = os.path.join(os.path.abspath('.'), script_path)

  headers[header_name] = script_path

def __handle_mode_option(_mode, request: Request, headers):
  headers[custom_headers.PROXY_MODE] = _mode

  if _mode == mode.MOCK or _mode == mode.TEST:
    # If mocking or testing, we already know which request to get response from 
    if request.id:
      headers[custom_headers.MOCK_REQUEST_ID] = str(request.id)

    headers[custom_headers.MOCK_POLICY] = mock_policy.ALL

def __create_replayed_response(request_id: int, res: requests.Response, latency: int):   
  replayed_response_model = ReplayedResponseModel(Settings.instance())
  raw = PythonResponseAdapterFactory(res).raw_response()
  return replayed_response_model.create(
    latency = latency,
    raw = raw,
    request_id = request_id,
    status = res.status_code,
  )

def __overwrite_response(replayed_response_id):
  replayed_response_model = ReplayedResponseModel(Settings.instance())
  return replayed_response_model.activate(replayed_response_id)


from typing import Union

from stoobly_agent.lib.utils.decode import decode

class RequestStringControl():
  REQUEST_TYPE = 1

  def __init__(self, c: Union[bytes, str] = None):
    if not c:
      self.__request_type = self.REQUEST_TYPE
      self.__id = None
      self.__timestamp = None # nanosecond
    else:
      self.parse(c)

  @property
  def id(self):
    return self.__id

  @id.setter
  def id(self, id: str):
    self.__id = id

  @property
  def request_type(self):
    return self.__request_type

  @property
  def timestamp(self):
    return self.__timestamp

  @timestamp.setter
  def timestamp(self, t: int):
    self.__timestamp = t

  def parse(self, s: Union[bytes, str]):
    toks = decode(s).split(' ')
    self.__id = toks[1]
    self.__timestamp = int(toks[2])

  def serialize(self):
    return "{} {} {}".format(self.REQUEST_TYPE, self.__id, self.__timestamp)